# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PySide import QtCore, QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP
import FreeCAD
import FreeCADGui
import PathScripts.PathIconViewProvider as PathIconViewProvider
import PathScripts.PathLog as PathLog
import PathScripts.PathPreferences as PathPreferences
import PathScripts.PathToolBit as PathToolBit
import PathScripts.PathToolBitEdit as PathToolBitEdit
import os

__title__ = "Tool Bit UI"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecadweb.org"
__doc__ = "Task panel editor for a ToolBit"


if False:
    PathLog.setLevel(PathLog.Level.DEBUG, PathLog.thisModule())
    PathLog.trackModule(PathLog.thisModule())
else:
    PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())

translate = FreeCAD.Qt.translate


class ViewProvider(object):
    """ViewProvider for a ToolBit.
    It's sole job is to provide an icon and invoke the TaskPanel on edit."""

    def __init__(self, vobj, name):
        PathLog.track(name, vobj.Object)
        self.panel = None
        self.icon = name
        self.obj = vobj.Object
        self.vobj = vobj
        vobj.Proxy = self

    def attach(self, vobj):
        PathLog.track(vobj.Object)
        self.vobj = vobj
        self.obj = vobj.Object

    def getIcon(self):
        png = self.obj.Proxy.getBitThumbnail(self.obj)
        if png:
            pixmap = QtGui.QPixmap()
            pixmap.loadFromData(png, "PNG")
            return QtGui.QIcon(pixmap)
        return ":/icons/Path_ToolBit.svg"

    def __getstate__(self):
        return None

    def __setstate__(self, state):
        return None

    def onDelete(self, vobj, arg2=None):
        PathLog.track(vobj.Object.Label)
        vobj.Object.Proxy.onDelete(vobj.Object)

    def getDisplayMode(self, mode):
        return "Default"

    def _openTaskPanel(self, vobj, deleteOnReject):
        PathLog.track()
        self.panel = TaskPanel(vobj, deleteOnReject)
        FreeCADGui.Control.closeDialog()
        FreeCADGui.Control.showDialog(self.panel)
        self.panel.setupUi()

    def setCreate(self, vobj):
        PathLog.track()
        self._openTaskPanel(vobj, True)

    def setEdit(self, vobj, mode=0):
        self._openTaskPanel(vobj, False)
        return True

    def unsetEdit(self, vobj, mode):
        FreeCADGui.Control.closeDialog()
        self.panel = None
        return

    def claimChildren(self):
        if self.obj.BitBody:
            return [self.obj.BitBody]
        return []

    def doubleClicked(self, vobj):
        if os.path.exists(vobj.Object.BitShape):
            self.setEdit(vobj)
        else:
            msg = translate(
                "PathToolBit", "Toolbit cannot be edited: Shapefile not found"
            )
            diag = QtGui.QMessageBox(QtGui.QMessageBox.Warning, "Error", msg)
            diag.setWindowModality(QtCore.Qt.ApplicationModal)
            diag.exec_()


class TaskPanel:
    """TaskPanel for the SetupSheet - if it is being edited directly."""

    def __init__(self, vobj, deleteOnReject):
        PathLog.track(vobj.Object.Label)
        self.vobj = vobj
        self.obj = vobj.Object
        self.editor = PathToolBitEdit.ToolBitEditor(self.obj)
        self.form = self.editor.form
        self.deleteOnReject = deleteOnReject
        FreeCAD.ActiveDocument.openTransaction("Edit ToolBit")

    def reject(self):
        FreeCAD.ActiveDocument.abortTransaction()
        self.editor.reject()
        FreeCADGui.Control.closeDialog()
        if self.deleteOnReject:
            FreeCAD.ActiveDocument.openTransaction("Uncreate ToolBit")
            self.editor.reject()
            FreeCAD.ActiveDocument.removeObject(self.obj.Name)
            FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()

    def accept(self):
        self.editor.accept()

        FreeCAD.ActiveDocument.commitTransaction()
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()

    def updateUI(self):
        PathLog.track()
        self.editor.updateUI()

    def updateModel(self):
        self.editor.updateTool()
        FreeCAD.ActiveDocument.recompute()

    def setupUi(self):
        self.editor.setupUI()


class ToolBitGuiFactory(PathToolBit.ToolBitFactory):
    def Create(self, name="ToolBit", shapeFile=None, path=None):
        """Create(name = 'ToolBit') ... creates a new tool bit.
        It is assumed the tool will be edited immediately so the internal bit body is still attached."""

        PathLog.track(name, shapeFile, path)
        FreeCAD.ActiveDocument.openTransaction("Create ToolBit")
        tool = PathToolBit.ToolBitFactory.Create(self, name, shapeFile, path)
        PathIconViewProvider.Attach(tool.ViewObject, name)
        FreeCAD.ActiveDocument.commitTransaction()
        return tool


def isValidFileName(filename):
    print(filename)
    try:
        with open(filename, "w") as tempfile:
            return True
    except Exception:
        return False


def GetNewToolFile(parent=None):
    if parent is None:
        parent = QtGui.QApplication.activeWindow()

    foo = QtGui.QFileDialog.getSaveFileName(
        parent, "Tool", PathPreferences.lastPathToolBit(), "*.fctb"
    )
    if foo and foo[0]:
        if not isValidFileName(foo[0]):
            msgBox = QtGui.QMessageBox()
            msg = translate("Path", "Invalid Filename")
            msgBox.setText(msg)
            msgBox.exec_()
        else:
            PathPreferences.setLastPathToolBit(os.path.dirname(foo[0]))
            return foo[0]
    return None


def GetToolFile(parent=None):
    if parent is None:
        parent = QtGui.QApplication.activeWindow()
    foo = QtGui.QFileDialog.getOpenFileName(
        parent, "Tool", PathPreferences.lastPathToolBit(), "*.fctb"
    )
    if foo and foo[0]:
        PathPreferences.setLastPathToolBit(os.path.dirname(foo[0]))
        return foo[0]
    return None


def GetToolFiles(parent=None):
    if parent is None:
        parent = QtGui.QApplication.activeWindow()
    foo = QtGui.QFileDialog.getOpenFileNames(
        parent, "Tool", PathPreferences.lastPathToolBit(), "*.fctb"
    )
    if foo and foo[0]:
        PathPreferences.setLastPathToolBit(os.path.dirname(foo[0][0]))
        return foo[0]
    return []


def GetToolShapeFile(parent=None):
    if parent is None:
        parent = QtGui.QApplication.activeWindow()

    location = PathPreferences.lastPathToolShape()
    if os.path.isfile(location):
        location = os.path.split(location)[0]
    elif not os.path.isdir(location):
        location = PathPreferences.filePath()

    fname = QtGui.QFileDialog.getOpenFileName(
        parent, "Select Tool Shape", location, "*.fcstd"
    )
    if fname and fname[0]:
        if fname != location:
            newloc = os.path.dirname(fname[0])
            PathPreferences.setLastPathToolShape(newloc)
        return fname[0]
    else:
        return None


def LoadTool(parent=None):
    """
    LoadTool(parent=None) ... Open a file dialog to load a tool from a file.
    """
    foo = GetToolFile(parent)
    return PathToolBit.Factory.CreateFrom(foo) if foo else foo


def LoadTools(parent=None):
    """
    LoadTool(parent=None) ... Open a file dialog to load a tool from a file.
    """
    return [PathToolBit.Factory.CreateFrom(foo) for foo in GetToolFiles(parent)]


# Set the factory so all tools are created with UI
PathToolBit.Factory = ToolBitGuiFactory()

PathIconViewProvider.RegisterViewProvider("ToolBit", ViewProvider)



# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Lesser General Public            *
# *   License as published by the Free Software Foundation; either          *
# *   version 2.1 of the License, or (at your option) any later version.    *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with this library; if not, write to the Free Software   *
# *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
#  *  02110-1301  USA                                                       *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
from PySide2 import QtCore, QtWidgets
import Addon

import os

translate = FreeCAD.Qt.translate


def ask_to_install_toolbar_button(repo: Addon) -> None:
    pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
    do_not_show_dialog = pref.GetBool("dontShowAddMacroButtonDialog", False)
    button_exists = check_for_button(repo)
    if not do_not_show_dialog and not button_exists:
        add_toolbar_button_dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "add_toolbar_button_dialog.ui")
        )
        add_toolbar_button_dialog.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)
        add_toolbar_button_dialog.buttonYes.clicked.connect(
            lambda: install_toolbar_button(repo)
        )
        add_toolbar_button_dialog.buttonNever.clicked.connect(
            lambda: pref.SetBool("dontShowAddMacroButtonDialog", True)
        )
        add_toolbar_button_dialog.exec()


def check_for_button(repo: Addon) -> bool:
    command = FreeCADGui.Command.findCustomCommand(repo.macro.filename)
    if not command:
        return False
    custom_toolbars = FreeCAD.ParamGet(
        "User parameter:BaseApp/Workbench/Global/Toolbar"
    )
    toolbar_groups = custom_toolbars.GetGroups()
    for group in toolbar_groups:
        toolbar = custom_toolbars.GetGroup(group)
        if toolbar.GetString(command, "*") != "*":
            return True
    return False


def ask_for_toolbar(
    repo: Addon, custom_toolbars
) -> object:  # Returns the pref group for the new toolbar
    pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")

    # In this one spot, default True: if this is the first time we got to
    # this chunk of code, we are always going to ask.
    ask = pref.GetBool("alwaysAskForToolbar", True)

    if ask:
        select_toolbar_dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "select_toolbar_dialog.ui")
        )
        select_toolbar_dialog.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)

        select_toolbar_dialog.comboBox.clear()

        for group in custom_toolbars:
            ref = FreeCAD.ParamGet(
                "User parameter:BaseApp/Workbench/Global/Toolbar/" + group
            )
            name = ref.GetString("Name", "")
            if name:
                select_toolbar_dialog.comboBox.addItem(name)
            else:
                FreeCAD.Console.PrintWarning(
                    f"Custom toolbar {group} does not have a Name element\n"
                )
        new_menubar_option_text = translate("AddonsInstaller", "Create new toolbar")
        select_toolbar_dialog.comboBox.addItem(new_menubar_option_text)

        result = select_toolbar_dialog.exec()
        if result == QtWidgets.QDialog.Accepted:
            selection = select_toolbar_dialog.comboBox.currentText()
            if select_toolbar_dialog.checkBox.checkState() == QtCore.Qt.Unchecked:
                pref.SetBool("alwaysAskForToolbar", False)
            else:
                pref.SetBool("alwaysAskForToolbar", True)
            if selection == new_menubar_option_text:
                return create_new_custom_toolbar()
            else:
                return get_toolbar_with_name(selection)
        else:
            return None
    else:
        custom_toolbar_name = pref.GetString(
            "CustomToolbarName", "Auto-Created Macro Toolbar"
        )
        toolbar = get_toolbar_with_name(custom_toolbar_name)
        if not toolbar:
            # They told us not to ask, but then the toolbar got deleted... ask anyway!
            ask = pref.RemBool("alwaysAskForToolbar")
            return ask_for_toolbar(repo, custom_toolbars)
        else:
            return toolbar


def get_toolbar_with_name(name: str) -> object:
    top_group = FreeCAD.ParamGet("User parameter:BaseApp/Workbench/Global/Toolbar")
    custom_toolbars = top_group.GetGroups()
    for toolbar in custom_toolbars:
        group = FreeCAD.ParamGet(
            "User parameter:BaseApp/Workbench/Global/Toolbar/" + toolbar
        )
        group_name = group.GetString("Name", "")
        if group_name == name:
            return group
    return None


def create_new_custom_toolbar() -> object:

    # We need two names: the name of the auto-created toolbar, as it will be displayed to the
    # user in various menus, and the underlying name of the toolbar group. Both must be
    # unique.

    # First, the displayed name
    custom_toolbar_name = "Auto-Created Macro Toolbar"
    pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
    top_group = FreeCAD.ParamGet("User parameter:BaseApp/Workbench/Global/Toolbar")
    custom_toolbars = top_group.GetGroups()
    name_taken = check_for_toolbar(custom_toolbars, custom_toolbar_name)
    if name_taken:
        i = 2  # Don't use (1), start at (2)
        while True:
            test_name = custom_toolbar_name + f" ({i})"
            if not check_for_toolbar(custom_toolbars, test_name):
                custom_toolbar_name = test_name
            i = i + 1

    # Second, the toolbar preference group name
    i = 1
    while True:
        new_group_name = "Custom_" + str(i)
        if new_group_name not in custom_toolbars:
            break
        i = i + 1

    custom_toolbar = FreeCAD.ParamGet(
        "User parameter:BaseApp/Workbench/Global/Toolbar/" + new_group_name
    )
    custom_toolbar.SetString("Name", custom_toolbar_name)
    custom_toolbar.SetBool("Active", True)
    return custom_toolbar


def check_for_toolbar(custom_toolbars, toolbar_name: str) -> bool:
    tb = get_toolbar_with_name(toolbar_name)
    if tb:
        return True
    else:
        return False


def install_toolbar_button(repo: Addon) -> None:
    pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
    custom_toolbar_name = pref.GetString(
        "CustomToolbarName", "Auto-Created Macro Toolbar"
    )

    # Default to false here: if the variable hasn't been set, we don't assume
    # that we have to ask, because the simplest is to just create a new toolbar
    # and never ask at all.
    ask = pref.GetBool("alwaysAskForToolbar", False)

    # See if there is already a custom toolbar for macros:
    top_group = FreeCAD.ParamGet("User parameter:BaseApp/Workbench/Global/Toolbar")
    custom_toolbars = top_group.GetGroups()
    if custom_toolbars:
        # If there are already custom toolbars, see if one of them is the one we used last time
        found_toolbar = False
        for toolbar_name in custom_toolbars:
            test_toolbar = FreeCAD.ParamGet(
                "User parameter:BaseApp/Workbench/Global/Toolbar/" + toolbar_name
            )
            name = test_toolbar.GetString("Name", "")
            if name == custom_toolbar_name:
                custom_toolbar = test_toolbar
                found_toolbar = True
                break
        if ask or not found_toolbar:
            # We have to ask the user what to do...
            custom_toolbar = ask_for_toolbar(repo, custom_toolbars)
            if custom_toolbar:
                custom_toolbar_name = custom_toolbar.GetString("Name")
                pref.SetString("CustomToolbarName", custom_toolbar_name)
    else:
        # Create a custom toolbar
        custom_toolbar = FreeCAD.ParamGet(
            "User parameter:BaseApp/Workbench/Global/Toolbar/Custom_1"
        )
        custom_toolbar.SetString("Name", custom_toolbar_name)
        custom_toolbar.SetBool("Active", True)

    if custom_toolbar:
        install_macro_to_toolbar(repo, custom_toolbar)
    else:
        FreeCAD.Console.PrintMessage(
            f"In the end, no custom toolbar was set, bailing out\n"
        )


def install_macro_to_toolbar(repo: Addon, toolbar: object) -> None:
    menuText = repo.display_name
    tooltipText = f"<b>{repo.display_name}</b>"
    if repo.macro.comment:
        tooltipText += f"<br/><p>{repo.macro.comment}</p>"
        whatsThisText = repo.macro.comment
    else:
        whatsThisText = translate(
            "AddonsInstaller", "A macro installed with the FreeCAD Addon Manager"
        )
    statustipText = (
        translate("AddonsInstaller", "Run", "Indicates a macro that can be 'run'")
        + " "
        + repo.display_name
    )
    if repo.macro.icon:
        if os.path.isabs(repo.macro.icon):
            pixmapText = os.path.normpath(repo.macro.icon)
        else:
            macro_repo_dir = FreeCAD.getUserMacroDir(True)
            pixmapText = os.path.normpath(os.path.join(macro_repo_dir, repo.macro.icon))
    elif repo.macro.xpm:
        macro_repo_dir = FreeCAD.getUserMacroDir(True)
        icon_file = os.path.normpath(
            os.path.join(macro_repo_dir, repo.macro.name + "_icon.xpm")
        )
        with open(icon_file, "w") as f:
            f.write(repo.macro.xpm)
        pixmapText = icon_file
    else:
        pixmapText = None

    # Add this command to that toolbar
    command_name = FreeCADGui.Command.createCustomCommand(
        repo.macro.filename,
        menuText,
        tooltipText,
        whatsThisText,
        statustipText,
        pixmapText,
    )
    toolbar.SetString(command_name, "FreeCAD")

    # Force the toolbars to be recreated
    wb = FreeCADGui.activeWorkbench()
    wb.reloadActive()


def remove_custom_toolbar_button(repo: Addon) -> None:
    """If this repo contains a macro, look through the custom commands and
    see if one is set up for this macro. If so, remove it, including any
    toolbar entries."""

    command = FreeCADGui.Command.findCustomCommand(repo.macro.filename)
    if not command:
        return
    custom_toolbars = FreeCAD.ParamGet(
        "User parameter:BaseApp/Workbench/Global/Toolbar"
    )
    toolbar_groups = custom_toolbars.GetGroups()
    for group in toolbar_groups:
        toolbar = custom_toolbars.GetGroup(group)
        if toolbar.GetString(command, "*") != "*":
            toolbar.RemString(command)

    FreeCADGui.Command.removeCustomCommand(command)

    # Force the toolbars to be recreated
    wb = FreeCADGui.activeWorkbench()
    wb.reloadActive()



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PySide import QtCore
from PySide.QtCore import QT_TRANSLATE_NOOP
import FreeCAD
import FreeCADGui
import PathScripts
import PathScripts.PathLog as PathLog
import os

if False:
    PathLog.setLevel(PathLog.Level.DEBUG, PathLog.thisModule())
    PathLog.trackModule(PathLog.thisModule())
else:
    PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())


class CommandToolBitCreate:
    """
    Command used to create a new Tool.
    """

    def __init__(self):
        pass

    def GetResources(self):
        return {
            "Pixmap": "Path_ToolBit",
            "MenuText": QT_TRANSLATE_NOOP("Path_ToolBitCreate", "Create Tool"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Path_ToolBitCreate", "Creates a new ToolBit object"
            ),
        }

    def IsActive(self):
        return FreeCAD.ActiveDocument is not None

    def Activated(self):
        obj = PathScripts.PathToolBit.Factory.Create()
        obj.ViewObject.Proxy.setCreate(obj.ViewObject)


class CommandToolBitSave:
    """
    Command used to save an existing Tool to a file.
    """

    def __init__(self, saveAs):
        self.saveAs = saveAs

    def GetResources(self):
        if self.saveAs:
            menuTxt = QT_TRANSLATE_NOOP("Path_ToolBitSaveAs", "Save Tool as...")
        else:
            menuTxt = QT_TRANSLATE_NOOP("Path_ToolBitSave", "Save Tool")
        return {
            "Pixmap": "Path_ToolBit",
            "MenuText": menuTxt,
            "ToolTip": QT_TRANSLATE_NOOP(
                "Path_ToolBitSave", "Save an existing ToolBit object to a file"
            ),
        }

    def selectedTool(self):
        sel = FreeCADGui.Selection.getSelectionEx()
        if 1 == len(sel) and isinstance(
            sel[0].Object.Proxy, PathScripts.PathToolBit.ToolBit
        ):
            return sel[0].Object
        return None

    def IsActive(self):
        tool = self.selectedTool()
        if tool:
            if tool.File:
                return True
            return self.saveAs
        return False

    def Activated(self):
        from PySide import QtGui

        tool = self.selectedTool()
        if tool:
            path = None
            if not tool.File or self.saveAs:
                if tool.File:
                    fname = tool.File
                else:
                    fname = os.path.join(
                        PathScripts.PathPreferences.lastPathToolBit(),
                        tool.Label + ".fctb",
                    )
                foo = QtGui.QFileDialog.getSaveFileName(
                    QtGui.QApplication.activeWindow(), "Tool", fname, "*.fctb"
                )
                if foo:
                    path = foo[0]
            else:
                path = tool.File

            if path:
                if not path.endswith(".fctb"):
                    path += ".fctb"
                tool.Proxy.saveToFile(tool, path)
                PathScripts.PathPreferences.setLastPathToolBit(os.path.dirname(path))


class CommandToolBitLoad:
    """
    Command used to load an existing Tool from a file into the current document.
    """

    def __init__(self):
        pass

    def GetResources(self):
        return {
            "Pixmap": "Path_ToolBit",
            "MenuText": QT_TRANSLATE_NOOP("Path_ToolBitLoad", "Load Tool"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Path_ToolBitLoad", "Load an existing ToolBit object from a file"
            ),
        }

    def selectedTool(self):
        sel = FreeCADGui.Selection.getSelectionEx()
        if 1 == len(sel) and isinstance(
            sel[0].Object.Proxy, PathScripts.PathToolBit.ToolBit
        ):
            return sel[0].Object
        return None

    def IsActive(self):
        return FreeCAD.ActiveDocument is not None

    def Activated(self):
        if PathScripts.PathToolBitGui.LoadTools():
            FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    FreeCADGui.addCommand("Path_ToolBitCreate", CommandToolBitCreate())
    FreeCADGui.addCommand("Path_ToolBitLoad", CommandToolBitLoad())
    FreeCADGui.addCommand("Path_ToolBitSave", CommandToolBitSave(False))
    FreeCADGui.addCommand("Path_ToolBitSaveAs", CommandToolBitSave(True))

CommandList = [
    "Path_ToolBitCreate",
    "Path_ToolBitLoad",
    "Path_ToolBitSave",
    "Path_ToolBitSaveAs",
]

FreeCAD.Console.PrintLog("Loading PathToolBitCmd... done\n")



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2016 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import PathScripts
import PathScripts.post
import PathScripts.PathProfileContour
import PathScripts.PathJob
import PathScripts.PathPost as PathPost
import PathScripts.PathToolController
import PathScripts.PathUtil
import PathScripts.PostUtils as PostUtils
import difflib
import unittest
import Path

WriteDebugOutput = False


class PathPostTestCases(unittest.TestCase):
    def setUp(self):
        testfile = FreeCAD.getHomePath() + "Mod/Path/PathTests/boxtest.fcstd"
        self.doc = FreeCAD.open(testfile)
        self.job = FreeCAD.ActiveDocument.getObject("Job")
        self.postlist = []
        currTool = None
        for obj in self.job.Group:
            if not isinstance(obj.Proxy, PathScripts.PathToolController.ToolController):
                tc = PathScripts.PathUtil.toolControllerForOp(obj)
                if tc is not None:
                    if tc.ToolNumber != currTool:
                        self.postlist.append(tc)
                self.postlist.append(obj)

    def tearDown(self):
        FreeCAD.closeDocument("boxtest")

    def testLinuxCNC(self):
        from PathScripts.post import linuxcnc_post as postprocessor

        args = (
            "--no-header --no-line-numbers --no-comments --no-show-editor --precision=2"
        )
        gcode = postprocessor.export(self.postlist, "gcode.tmp", args)

        referenceFile = (
            FreeCAD.getHomePath() + "Mod/Path/PathTests/test_linuxcnc_00.ngc"
        )
        with open(referenceFile, "r") as fp:
            refGCode = fp.read()

        # Use if this test fails in order to have a real good look at the changes
        if WriteDebugOutput:
            with open("testLinuxCNC.tmp", "w") as fp:
                fp.write(gcode)

        if gcode != refGCode:
            msg = "".join(
                difflib.ndiff(gcode.splitlines(True), refGCode.splitlines(True))
            )
            self.fail("linuxcnc output doesn't match: " + msg)

    def testLinuxCNCImperial(self):
        from PathScripts.post import linuxcnc_post as postprocessor

        args = "--no-header --no-line-numbers --no-comments --no-show-editor --precision=2 --inches"
        gcode = postprocessor.export(self.postlist, "gcode.tmp", args)

        referenceFile = (
            FreeCAD.getHomePath() + "Mod/Path/PathTests/test_linuxcnc_10.ngc"
        )
        with open(referenceFile, "r") as fp:
            refGCode = fp.read()

        # Use if this test fails in order to have a real good look at the changes
        if WriteDebugOutput:
            with open("testLinuxCNCImplerial.tmp", "w") as fp:
                fp.write(gcode)

        if gcode != refGCode:
            msg = "".join(
                difflib.ndiff(gcode.splitlines(True), refGCode.splitlines(True))
            )
            self.fail("linuxcnc output doesn't match: " + msg)

    def testCentroid(self):
        from PathScripts.post import centroid_post as postprocessor

        args = "--no-header --no-line-numbers --no-comments --no-show-editor --axis-precision=2 --feed-precision=2"
        gcode = postprocessor.export(self.postlist, "gcode.tmp", args)

        referenceFile = (
            FreeCAD.getHomePath() + "Mod/Path/PathTests/test_centroid_00.ngc"
        )
        with open(referenceFile, "r") as fp:
            refGCode = fp.read()

        # Use if this test fails in order to have a real good look at the changes
        if WriteDebugOutput:
            with open("testCentroid.tmp", "w") as fp:
                fp.write(gcode)

        if gcode != refGCode:
            msg = "".join(
                difflib.ndiff(gcode.splitlines(True), refGCode.splitlines(True))
            )
            self.fail("linuxcnc output doesn't match: " + msg)


class TestPathPostUtils(unittest.TestCase):
    def testSplitArcs(self):

        commands = [
            Path.Command("G1 X-7.5 Y5.0 Z0.0"),
            Path.Command("G2 I2.5 J0.0 K0.0 X-5.0 Y7.5 Z0.0"),
            Path.Command("G1 X5.0 Y7.5 Z0.0"),
            Path.Command("G2 I0.0 J-2.5 K0.0 X7.5 Y5.0 Z0.0"),
            Path.Command("G1 X7.5 Y-5.0 Z0.0"),
            Path.Command("G2 I-2.5 J0.0 K0.0 X5.0 Y-7.5 Z0.0"),
            Path.Command("G1 X-5.0 Y-7.5 Z0.0"),
            Path.Command("G2 I0.0 J2.5 K0.0 X-7.5 Y-5.0 Z0.0"),
            Path.Command("G1 X-7.5 Y0.0 Z0.0"),
        ]

        testpath = Path.Path(commands)
        self.assertTrue(len(testpath.Commands) == 9)
        self.assertTrue(
            len([c for c in testpath.Commands if c.Name in ["G2", "G3"]]) == 4
        )

        results = PostUtils.splitArcs(testpath)
        # self.assertTrue(len(results.Commands) == 117)
        self.assertTrue(
            len([c for c in results.Commands if c.Name in ["G2", "G3"]]) == 0
        )


class TestPathPostImport(unittest.TestCase):
    def test001(self):
        """test001()... Verify 'No active document' exception thrown if no document open."""
        from PathScripts.post import gcode_pre as gcode_pre

        self.assertRaises(
            gcode_pre.PathNoActiveDocumentException,
            gcode_pre._isImportEnvironmentReady,
        )

    def test002(self):
        """test002()... Verify 'No job object' exception thrown if Job object available."""
        from PathScripts.post import gcode_pre as gcode_pre

        doc = FreeCAD.newDocument("TestPathPost")

        self.assertRaises(
            gcode_pre.PathNoJobException,
            gcode_pre._isImportEnvironmentReady,
        )
        FreeCAD.closeDocument(doc.Name)

    def test003(self, close=True):
        """test003()... Verify 'No job object' exception thrown if Job object available."""
        from PathScripts.post import gcode_pre as gcode_pre

        doc = FreeCAD.newDocument("TestPathPost")

        # Add temporary receiving Job object
        box = FreeCAD.ActiveDocument.addObject("Part::Box", "Box")
        box.Label = "Temporary Box"
        # Add Job object with view provider support when possible
        if FreeCAD.GuiUp:
            import PathScripts.PathJobGui as PathJobGui

            box.ViewObject.Visibility = False
            job = PathJobGui.Create([box], openTaskPanel=False)
        else:
            import PathScripts.PathJob as PathJob

            job = PathJob.Create("Job", [box])

        importFile = FreeCAD.getHomePath() + "Mod/Path/PathTests/test_centroid_00.ngc"
        gcode_pre.insert(importFile, "test_centroid_00")

        # self.assertTrue(doc.Name == "test_centroid_00")

        opList = doc.Job.Operations.Group
        self.assertTrue(
            len(opList) == 2,
            "Expected 2 Custom operations to be created from source g-code file, test_centroid_00.ngc",
        )
        self.assertTrue(
            opList[0].Name == "Custom", "Expected first operation to be Custom"
        )
        self.assertTrue(
            opList[1].Name == "Custom001", "Expected second operation to be Custom001"
        )

        if close:
            FreeCAD.closeDocument(doc.Name)

    def test004(self):
        """test004()... Verify g-code imported with g-code pre-processor"""

        self.test003(close=False)

        doc = FreeCAD.ActiveDocument
        op1 = doc.Job.Operations.Group[0]
        op2 = doc.Job.Operations.Group[1]

        # Verify g-code sizes
        self.assertTrue(
            op1.Path.Size == 4, "Expected Custom g-code command count to be 4."
        )
        self.assertTrue(
            op2.Path.Size == 60, "Expected Custom g-code command count to be 60."
        )

        # Verify g-code commands
        op1_code = (
            "(Custom_test_centroid_00)\n(Begin Custom)\nG90 G49.000000\n(End Custom)\n"
        )
        op2_code = "(Custom001_test_centroid_00)\n(Begin Custom)\nG0 Z15.000000\nG90\nG0 Z15.000000\nG0 X10.000000 Y10.000000\nG0 Z10.000000\nG1 X10.000000 Y10.000000 Z9.000000\nG1 X10.000000 Y0.000000 Z9.000000\nG1 X0.000000 Y0.000000 Z9.000000\nG1 X0.000000 Y10.000000 Z9.000000\nG1 X10.000000 Y10.000000 Z9.000000\nG1 X10.000000 Y10.000000 Z8.000000\nG1 X10.000000 Y0.000000 Z8.000000\nG1 X0.000000 Y0.000000 Z8.000000\nG1 X0.000000 Y10.000000 Z8.000000\nG1 X10.000000 Y10.000000 Z8.000000\nG1 X10.000000 Y10.000000 Z7.000000\nG1 X10.000000 Y0.000000 Z7.000000\nG1 X0.000000 Y0.000000 Z7.000000\nG1 X0.000000 Y10.000000 Z7.000000\nG1 X10.000000 Y10.000000 Z7.000000\nG1 X10.000000 Y10.000000 Z6.000000\nG1 X10.000000 Y0.000000 Z6.000000\nG1 X0.000000 Y0.000000 Z6.000000\nG1 X0.000000 Y10.000000 Z6.000000\nG1 X10.000000 Y10.000000 Z6.000000\nG1 X10.000000 Y10.000000 Z5.000000\nG1 X10.000000 Y0.000000 Z5.000000\nG1 X0.000000 Y0.000000 Z5.000000\nG1 X0.000000 Y10.000000 Z5.000000\nG1 X10.000000 Y10.000000 Z5.000000\nG1 X10.000000 Y10.000000 Z4.000000\nG1 X10.000000 Y0.000000 Z4.000000\nG1 X0.000000 Y0.000000 Z4.000000\nG1 X0.000000 Y10.000000 Z4.000000\nG1 X10.000000 Y10.000000 Z4.000000\nG1 X10.000000 Y10.000000 Z3.000000\nG1 X10.000000 Y0.000000 Z3.000000\nG1 X0.000000 Y0.000000 Z3.000000\nG1 X0.000000 Y10.000000 Z3.000000\nG1 X10.000000 Y10.000000 Z3.000000\nG1 X10.000000 Y10.000000 Z2.000000\nG1 X10.000000 Y0.000000 Z2.000000\nG1 X0.000000 Y0.000000 Z2.000000\nG1 X0.000000 Y10.000000 Z2.000000\nG1 X10.000000 Y10.000000 Z2.000000\nG1 X10.000000 Y10.000000 Z1.000000\nG1 X10.000000 Y0.000000 Z1.000000\nG1 X0.000000 Y0.000000 Z1.000000\nG1 X0.000000 Y10.000000 Z1.000000\nG1 X10.000000 Y10.000000 Z1.000000\nG1 X10.000000 Y10.000000 Z0.000000\nG1 X10.000000 Y0.000000 Z0.000000\nG1 X0.000000 Y0.000000 Z0.000000\nG1 X0.000000 Y10.000000 Z0.000000\nG1 X10.000000 Y10.000000 Z0.000000\nG0 Z15.000000\nG90 G49.000000\n(End Custom)\n"
        code1 = op1.Path.toGCode()
        self.assertTrue(
            code1 == op1_code,
            f"Gcode is not what is expected:\n~~~\n{code1}\n~~~\n\n\n~~~\n{op1_code}\n~~~",
        )
        code2 = op2.Path.toGCode()
        self.assertTrue(
            code2 == op2_code,
            f"Gcode is not what is expected:\n~~~\n{code2}\n~~~\n\n\n~~~\n{op2_code}\n~~~",
        )
        FreeCAD.closeDocument(doc.Name)

    def test005(self):
        """test005()... verify `_identifygcodeByToolNumberList()` produces correct output"""

        from PathScripts.post import gcode_pre as gcode_pre

        importFile = FreeCAD.getHomePath() + "Mod/Path/PathTests/test_centroid_00.ngc"
        gcodeByToolNumberList = gcode_pre._identifygcodeByToolNumberList(importFile)

        self.assertTrue(gcodeByToolNumberList[0][0] == ["G90 G80 G40 G49"])
        self.assertTrue(gcodeByToolNumberList[0][1] == 0)

        self.assertTrue(
            gcodeByToolNumberList[1][0]
            == [
                "G0 Z15.00",
                "G90",
                "G0 Z15.00",
                "G0 X10.00 Y10.00",
                "G0 Z10.00",
                "G1 X10.00 Y10.00 Z9.00",
                "G1 X10.00 Y0.00 Z9.00",
                "G1 X0.00 Y0.00 Z9.00",
                "G1 X0.00 Y10.00 Z9.00",
                "G1 X10.00 Y10.00 Z9.00",
                "G1 X10.00 Y10.00 Z8.00",
                "G1 X10.00 Y0.00 Z8.00",
                "G1 X0.00 Y0.00 Z8.00",
                "G1 X0.00 Y10.00 Z8.00",
                "G1 X10.00 Y10.00 Z8.00",
                "G1 X10.00 Y10.00 Z7.00",
                "G1 X10.00 Y0.00 Z7.00",
                "G1 X0.00 Y0.00 Z7.00",
                "G1 X0.00 Y10.00 Z7.00",
                "G1 X10.00 Y10.00 Z7.00",
                "G1 X10.00 Y10.00 Z6.00",
                "G1 X10.00 Y0.00 Z6.00",
                "G1 X0.00 Y0.00 Z6.00",
                "G1 X0.00 Y10.00 Z6.00",
                "G1 X10.00 Y10.00 Z6.00",
                "G1 X10.00 Y10.00 Z5.00",
                "G1 X10.00 Y0.00 Z5.00",
                "G1 X0.00 Y0.00 Z5.00",
                "G1 X0.00 Y10.00 Z5.00",
                "G1 X10.00 Y10.00 Z5.00",
                "G1 X10.00 Y10.00 Z4.00",
                "G1 X10.00 Y0.00 Z4.00",
                "G1 X0.00 Y0.00 Z4.00",
                "G1 X0.00 Y10.00 Z4.00",
                "G1 X10.00 Y10.00 Z4.00",
                "G1 X10.00 Y10.00 Z3.00",
                "G1 X10.00 Y0.00 Z3.00",
                "G1 X0.00 Y0.00 Z3.00",
                "G1 X0.00 Y10.00 Z3.00",
                "G1 X10.00 Y10.00 Z3.00",
                "G1 X10.00 Y10.00 Z2.00",
                "G1 X10.00 Y0.00 Z2.00",
                "G1 X0.00 Y0.00 Z2.00",
                "G1 X0.00 Y10.00 Z2.00",
                "G1 X10.00 Y10.00 Z2.00",
                "G1 X10.00 Y10.00 Z1.00",
                "G1 X10.00 Y0.00 Z1.00",
                "G1 X0.00 Y0.00 Z1.00",
                "G1 X0.00 Y10.00 Z1.00",
                "G1 X10.00 Y10.00 Z1.00",
                "G1 X10.00 Y10.00 Z0.00",
                "G1 X10.00 Y0.00 Z0.00",
                "G1 X0.00 Y0.00 Z0.00",
                "G1 X0.00 Y10.00 Z0.00",
                "G1 X10.00 Y10.00 Z0.00",
                "G0 Z15.00",
                "G90 G80 G40 G49",
            ]
        )
        self.assertTrue(gcodeByToolNumberList[1][1] == 2)


class OutputOrderingTestCases(unittest.TestCase):
    def setUp(self):
        testfile = FreeCAD.getHomePath() + "Mod/Path/PathTests/boxtest.fcstd"
        self.doc = FreeCAD.open(testfile)
        self.job = FreeCAD.ActiveDocument.getObject("Job001")

    def tearDown(self):
        FreeCAD.closeDocument("boxtest")

    def test010(self):
        # Basic postprocessing:

        self.job.Fixtures = ["G54"]
        self.job.SplitOutput = False
        self.job.OrderOutputBy = "Fixture"

        cpp = PathPost.CommandPathPost
        self.postlist = cpp.buildPostList(self, self.job)

        outlist = [i.Label for i in self.postlist[0]]

        self.assertTrue(len(self.postlist) == 1)
        expected = [
            "G54",
            "T1",
            "FirstOp-(T1)",
            "SecondOp-(T1)",
            "T2",
            "ThirdOp-(T2)",
            "T1",
            "FourthOp-(T1)",
            "T3",
            "FifthOp-(T3)",
        ]
        self.assertListEqual(outlist, expected)

    def test020(self):
        # Multiple Fixtures

        self.job.Fixtures = ["G54", "G55"]
        self.job.SplitOutput = False
        self.job.OrderOutputBy = "Fixture"

        cpp = PathPost.CommandPathPost
        self.postlist = cpp.buildPostList(self, self.job)

        self.assertTrue(len(self.postlist) == 1)

        outlist = [i.Label for i in self.postlist[0]]
        expected = [
            "G54",
            "T1",
            "FirstOp-(T1)",
            "SecondOp-(T1)",
            "T2",
            "ThirdOp-(T2)",
            "T1",
            "FourthOp-(T1)",
            "T3",
            "FifthOp-(T3)",
            "G55",
            "T1",
            "FirstOp-(T1)",
            "SecondOp-(T1)",
            "T2",
            "ThirdOp-(T2)",
            "T1",
            "FourthOp-(T1)",
            "T3",
            "FifthOp-(T3)",
        ]

        self.assertListEqual(outlist, expected)

    def test030(self):
        # Multiple Fixtures - Split output

        self.job.Fixtures = ["G54", "G55"]
        self.job.SplitOutput = True
        self.job.OrderOutputBy = "Fixture"

        cpp = PathPost.CommandPathPost
        self.postlist = cpp.buildPostList(self, self.job)

        self.assertTrue(len(self.postlist) == 2)

        outlist = [i.Label for i in self.postlist[0]]
        print(outlist)

        expected = [
            "G54",
            "T1",
            "FirstOp-(T1)",
            "SecondOp-(T1)",
            "T2",
            "ThirdOp-(T2)",
            "T1",
            "FourthOp-(T1)",
            "T3",
            "FifthOp-(T3)",
        ]
        self.assertListEqual(outlist, expected)

        expected = [
            "G55",
            "T1",
            "FirstOp-(T1)",
            "SecondOp-(T1)",
            "T2",
            "ThirdOp-(T2)",
            "T1",
            "FourthOp-(T1)",
            "T3",
            "FifthOp-(T3)",
        ]
        outlist = [i.Label for i in self.postlist[1]]
        self.assertListEqual(outlist, expected)

    def test040(self):
        # Order by 'Tool'

        self.job.Fixtures = ["G54", "G55"]
        self.job.SplitOutput = False
        self.job.OrderOutputBy = "Tool"

        cpp = PathPost.CommandPathPost
        self.postlist = cpp.buildPostList(self, self.job)
        outlist = [i.Label for i in self.postlist[0]]

        self.assertTrue(len(self.postlist) == 1)
        expected = [
            "G54",
            "T1",
            "FirstOp-(T1)",
            "SecondOp-(T1)",
            "T2",
            "ThirdOp-(T2)",
            "T1",
            "FourthOp-(T1)",
            "G55",
            "FirstOp-(T1)",
            "SecondOp-(T1)",
            "T2",
            "ThirdOp-(T2)",
            "T1",
            "FourthOp-(T1)",
            "G54",
            "T3",
            "FifthOp-(T3)",
            "G55",
            "FifthOp-(T3)",
        ]

        self.assertListEqual(outlist, expected)

    def test050(self):
        # Order by 'Tool' and split

        self.job.Fixtures = ["G54", "G55"]
        self.job.SplitOutput = True
        self.job.OrderOutputBy = "Tool"

        cpp = PathPost.CommandPathPost
        self.postlist = cpp.buildPostList(self, self.job)
        outlist = [i.Label for i in self.postlist[0]]

        expected = [
            "G54",
            "T1",
            "FirstOp-(T1)",
            "SecondOp-(T1)",
            "T2",
            "ThirdOp-(T2)",
            "T1",
            "FourthOp-(T1)",
            "G55",
            "FirstOp-(T1)",
            "SecondOp-(T1)",
            "T2",
            "ThirdOp-(T2)",
            "T1",
            "FourthOp-(T1)",
        ]
        self.assertListEqual(outlist, expected)

        outlist = [i.Label for i in self.postlist[1]]

        expected = [
            "G54",
            "T3",
            "FifthOp-(T3)",
            "G55",
            "FifthOp-(T3)",
        ]
        self.assertListEqual(outlist, expected)

    def test060(self):
        # Order by 'Operation'

        self.job.Fixtures = ["G54", "G55"]
        self.job.SplitOutput = False
        self.job.OrderOutputBy = "Operation"

        cpp = PathPost.CommandPathPost
        self.postlist = cpp.buildPostList(self, self.job)
        outlist = [i.Label for i in self.postlist[0]]

        self.assertTrue(len(self.postlist) == 1)
        expected = [
            "G54",
            "T1",
            "FirstOp-(T1)",
            "G55",
            "FirstOp-(T1)",
            "G54",
            "T1",
            "SecondOp-(T1)",
            "G55",
            "SecondOp-(T1)",
            "G54",
            "T2",
            "ThirdOp-(T2)",
            "G55",
            "ThirdOp-(T2)",
            "G54",
            "T1",
            "FourthOp-(T1)",
            "G55",
            "FourthOp-(T1)",
            "G54",
            "T3",
            "FifthOp-(T3)",
            "G55",
            "FifthOp-(T3)",
        ]

        self.assertListEqual(outlist, expected)

    def test070(self):
        # Order by 'Operation' and split

        self.job.Fixtures = ["G54", "G55"]
        self.job.SplitOutput = True
        self.job.OrderOutputBy = "Operation"

        cpp = PathPost.CommandPathPost
        self.postlist = cpp.buildPostList(self, self.job)
        self.assertTrue(len(self.postlist) == 5)

        outlist = [i.Label for i in self.postlist[0]]
        expected = [
            "G54",
            "T1",
            "FirstOp-(T1)",
            "G55",
            "FirstOp-(T1)",
        ]
        self.assertListEqual(outlist, expected)

        outlist = [i.Label for i in self.postlist[1]]
        expected = [
            "G54",
            "T1",
            "SecondOp-(T1)",
            "G55",
            "SecondOp-(T1)",
        ]
        self.assertListEqual(outlist, expected)

        outlist = [i.Label for i in self.postlist[2]]
        expected = [
            "G54",
            "T2",
            "ThirdOp-(T2)",
            "G55",
            "ThirdOp-(T2)",
        ]
        self.assertListEqual(outlist, expected)

        outlist = [i.Label for i in self.postlist[3]]
        expected = [
            "G54",
            "T1",
            "FourthOp-(T1)",
            "G55",
            "FourthOp-(T1)",
        ]
        self.assertListEqual(outlist, expected)

        outlist = [i.Label for i in self.postlist[4]]
        expected = [
            "G54",
            "T3",
            "FifthOp-(T3)",
            "G55",
            "FifthOp-(T3)",
        ]
        self.assertListEqual(outlist, expected)



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint contact"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecadweb.org"


def get_analysis_types():
    return "all"    # write for all analysis types


def get_sets_name():
    return "constraints_contact_surface_sets"


def get_constraint_title():
    return "Contact Constraints"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def get_before_write_constraint():
    return ""


def get_after_write_constraint():
    return ""


def write_meshdata_constraint(f, femobj, contact_obj, ccxwriter):
    # slave DEP
    f.write("*SURFACE, NAME=DEP{}\n".format(contact_obj.Name))
    for i in femobj["ContactSlaveFaces"]:
        f.write("{},S{}\n".format(i[0], i[1]))
    # master IND
    f.write("*SURFACE, NAME=IND{}\n".format(contact_obj.Name))
    for i in femobj["ContactMasterFaces"]:
        f.write("{},S{}\n".format(i[0], i[1]))


def write_constraint(f, femobj, contact_obj, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module

    f.write(
        "*CONTACT PAIR, INTERACTION=INT{},TYPE=SURFACE TO SURFACE\n"
        .format(contact_obj.Name)
    )
    ind_surf = "IND" + contact_obj.Name
    dep_surf = "DEP" + contact_obj.Name
    f.write("{},{}\n".format(dep_surf, ind_surf))
    f.write("*SURFACE INTERACTION, NAME=INT{}\n".format(contact_obj.Name))
    f.write("*SURFACE BEHAVIOR,PRESSURE-OVERCLOSURE=LINEAR\n")
    slope = contact_obj.Slope
    f.write("{:.13G}\n".format(slope))
    friction = contact_obj.Friction
    if friction > 0:
        f.write("*FRICTION \n")
        stick = (slope / 10.0)
        f.write("{:.13G}, {:.13G}\n".format(friction, stick))



# ***************************************************************************
# *   Copyright (c) 2021 Tobias Vaara <t@vaara.se>                          *
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD

import Fem
import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc


def get_information():
    return {
        "name": "Flexural Buckling",
        "meshtype": "solid",
        "meshelement": "Hexa8",
        "constraints": ["fixed", "force"],
        "solvers": ["calculix", "ccxtools"],
        "material": "solid",
        "equation": "buckling"
    }


def get_explanation(header=""):
    return header + """

To run the example from Python console use:
from femexamples.ccx_buckling_flexuralbuckling import setup
setup()


See forum topic post:
https://forum.freecadweb.org/viewtopic.php?f=18&t=20217&start=80#p488666

This example is based on a CalculiX verification example.
http://www.feacluster.com/CalculiX/ccx_2.13/doc/ccx/input_deck_viewer.php?input_deck=beam8b.inp

"""


def setup(doc=None, solvertype="ccxtools"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric object
    geom_obj = doc.addObject("Part::Box", "Beam")
    geom_obj.Length = 1
    geom_obj.Width = 1.5
    geom_obj.Height = 8
    doc.recompute()
    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver,
    if solvertype == "calculix":
        solver_obj = ObjectsFem.makeSolverCalculix(doc, "SolverCalculiX")
    elif solvertype == "ccxtools":
        solver_obj = ObjectsFem.makeSolverCalculixCcxTools(doc, "CalculiXccxTools")
        solver_obj.WorkingDir = u""
    else:
        FreeCAD.Console.PrintWarning(
            "Not known or not supported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    if solvertype == "calculix" or solvertype == "ccxtools":
        solver_obj.SplitInputWriter = False
        solver_obj.AnalysisType = "buckling"
        solver_obj.BucklingFactors = 10
        solver_obj.GeometricalNonlinearity = "linear"
        solver_obj.ThermoMechSteadyState = False
        solver_obj.MatrixSolverType = "default"
        solver_obj.IterationsControlParameterTimeUse = False
    analysis.addObject(solver_obj)

    # material
    material_obj = ObjectsFem.makeMaterialSolid(doc, "MechanicalMaterial")
    mat = material_obj.Material
    mat["Name"] = "CalculiX-Steel"
    mat["YoungsModulus"] = "210000 MPa"
    mat["PoissonRatio"] = "0.30"
    material_obj.Material = mat
    analysis.addObject(material_obj)

    # constraint fixed
    con_fixed = ObjectsFem.makeConstraintFixed(doc, "ConstraintFixed")
    con_fixed.References = [(geom_obj, "Face5")]
    analysis.addObject(con_fixed)

    # constraint force
    con_force = ObjectsFem.makeConstraintForce(doc, "ConstraintForce")
    con_force.References = [(geom_obj, "Face6")]
    con_force.Force = 21
    con_force.Reversed = True
    analysis.addObject(con_force)

    # mesh
    from .meshes.mesh_flexural_buckling import create_nodes, create_elements
    fem_mesh = Fem.FemMesh()
    control = create_nodes(fem_mesh)
    if not control:
        FreeCAD.Console.PrintError("Error on creating nodes.\n")
    control = create_elements(fem_mesh)
    if not control:
        FreeCAD.Console.PrintError("Error on creating elements.\n")
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.FemMesh = fem_mesh
    femmesh_obj.Part = geom_obj

    doc.recompute()
    return doc



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'compact_view.ui'
##
## Created by: Qt User Interface Compiler version 5.15.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_CompactView(object):
    def setupUi(self, CompactView):
        if not CompactView.objectName():
            CompactView.setObjectName(u"CompactView")
        CompactView.resize(489, 16)
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(CompactView.sizePolicy().hasHeightForWidth())
        CompactView.setSizePolicy(sizePolicy)
        self.horizontalLayout_2 = QHBoxLayout(CompactView)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setSizeConstraint(QLayout.SetNoConstraint)
        self.horizontalLayout_2.setContentsMargins(3, 0, 9, 0)
        self.labelIcon = QLabel(CompactView)
        self.labelIcon.setObjectName(u"labelIcon")
        sizePolicy1 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.labelIcon.sizePolicy().hasHeightForWidth())
        self.labelIcon.setSizePolicy(sizePolicy1)
        self.labelIcon.setMinimumSize(QSize(16, 16))
        self.labelIcon.setBaseSize(QSize(16, 16))

        self.horizontalLayout_2.addWidget(self.labelIcon)

        self.labelPackageName = QLabel(CompactView)
        self.labelPackageName.setObjectName(u"labelPackageName")

        self.horizontalLayout_2.addWidget(self.labelPackageName)

        self.labelVersion = QLabel(CompactView)
        self.labelVersion.setObjectName(u"labelVersion")

        self.horizontalLayout_2.addWidget(self.labelVersion)

        self.labelDescription = QLabel(CompactView)
        self.labelDescription.setObjectName(u"labelDescription")
        sizePolicy2 = QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Preferred)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(
            self.labelDescription.sizePolicy().hasHeightForWidth()
        )
        self.labelDescription.setSizePolicy(sizePolicy2)
        self.labelDescription.setTextFormat(Qt.PlainText)
        self.labelDescription.setWordWrap(False)
        self.labelDescription.setTextInteractionFlags(Qt.TextSelectableByMouse)

        self.horizontalLayout_2.addWidget(self.labelDescription)

        self.labelStatus = QLabel(CompactView)
        self.labelStatus.setObjectName(u"labelStatus")

        self.horizontalLayout_2.addWidget(self.labelStatus)

        self.retranslateUi(CompactView)

        QMetaObject.connectSlotsByName(CompactView)

    # setupUi

    def retranslateUi(self, CompactView):
        CompactView.setWindowTitle(
            QCoreApplication.translate("CompactView", u"Form", None)
        )
        self.labelIcon.setText(QCoreApplication.translate("CompactView", u"Icon", None))
        self.labelPackageName.setText(
            QCoreApplication.translate("CompactView", u"<b>Package Name</b>", None)
        )
        self.labelVersion.setText(
            QCoreApplication.translate("CompactView", u"Version", None)
        )
        self.labelDescription.setText(
            QCoreApplication.translate("CompactView", u"Description", None)
        )
        self.labelStatus.setText(
            QCoreApplication.translate("CompactView", u"UpdateAvailable", None)
        )

    # retranslateUi



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2015 Dan Falck <ddfalck@gmail.com>                      *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

""" Post Process command that will make use of the Output File and Post Processor entries in PathJob """

from __future__ import print_function

import FreeCAD
import FreeCADGui
import Path
import PathScripts.PathJob as PathJob
import PathScripts.PathLog as PathLog
import PathScripts.PathPreferences as PathPreferences
import PathScripts.PathUtil as PathUtil
import PathScripts.PathUtils as PathUtils
import os

from PathScripts.PathPostProcessor import PostProcessor
from PySide import QtCore, QtGui
from datetime import datetime
from PySide.QtCore import QT_TRANSLATE_NOOP

LOG_MODULE = PathLog.thisModule()

if False:
    PathLog.setLevel(PathLog.Level.DEBUG, PathLog.thisModule())
    PathLog.trackModule(PathLog.thisModule())
else:
    PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())

translate = FreeCAD.Qt.translate


class _TempObject:
    Path = None
    Name = "Fixture"
    Label = "Fixture"


class DlgSelectPostProcessor:
    def __init__(self, parent=None):
        self.dialog = FreeCADGui.PySideUic.loadUi(":/panels/DlgSelectPostProcessor.ui")
        firstItem = None
        for post in PathPreferences.allEnabledPostProcessors():
            item = QtGui.QListWidgetItem(post)
            item.setFlags(
                QtCore.Qt.ItemFlag.ItemIsSelectable | QtCore.Qt.ItemFlag.ItemIsEnabled
            )
            self.dialog.lwPostProcessor.addItem(item)
            if not firstItem:
                firstItem = item
        if firstItem:
            self.dialog.lwPostProcessor.setCurrentItem(firstItem)
        else:
            self.dialog.buttonBox.button(QtGui.QDialogButtonBox.Ok).setEnabled(False)
        self.tooltips = {}
        self.dialog.lwPostProcessor.itemDoubleClicked.connect(self.dialog.accept)
        self.dialog.lwPostProcessor.setMouseTracking(True)
        self.dialog.lwPostProcessor.itemEntered.connect(self.updateTooltip)

    def updateTooltip(self, item):
        if item.text() in self.tooltips.keys():
            tooltip = self.tooltips[item.text()]
        else:
            processor = PostProcessor.load(item.text())
            self.tooltips[item.text()] = processor.tooltip
            tooltip = processor.tooltip
        self.dialog.lwPostProcessor.setToolTip(tooltip)

    def exec_(self):
        if self.dialog.exec_() == 1:
            posts = self.dialog.lwPostProcessor.selectedItems()
            return posts[0].text()
        return None


class CommandPathPost:
    subpart = 1

    def resolveFileName(self, job):
        path = PathPreferences.defaultOutputFile()
        if job.PostProcessorOutputFile:
            path = job.PostProcessorOutputFile
        filename = path

        if "%D" in filename:
            D = FreeCAD.ActiveDocument.FileName
            if D:
                D = os.path.dirname(D)
                # in case the document is in the current working directory
                if not D:
                    D = "."
            else:
                FreeCAD.Console.PrintError(
                    "Please save document in order to resolve output path!\n"
                )
                return None
            filename = filename.replace("%D", D)

        if "%d" in filename:
            d = FreeCAD.ActiveDocument.Label
            filename = filename.replace("%d", d)

        if "%j" in filename:
            j = job.Label
            filename = filename.replace("%j", j)

        if "%M" in filename:
            pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Macro")
            M = pref.GetString("MacroPath", FreeCAD.getUserAppDataDir())
            filename = filename.replace("%M", M)

        if "%s" in filename:
            if job.SplitOutput:
                filename = filename.replace("%s", "_" + str(self.subpart))
                self.subpart += 1
            else:
                filename = filename.replace("%s", "")

        policy = PathPreferences.defaultOutputPolicy()

        openDialog = policy == "Open File Dialog"
        if os.path.isdir(filename) or not os.path.isdir(os.path.dirname(filename)):
            # Either the entire filename resolves into a directory or the parent directory doesn't exist.
            # Either way I don't know what to do - ask for help
            openDialog = True

        if os.path.isfile(filename) and not openDialog:
            if policy == "Open File Dialog on conflict":
                openDialog = True
            elif policy == "Append Unique ID on conflict":
                fn, ext = os.path.splitext(filename)
                nr = fn[-3:]
                n = 1
                if nr.isdigit():
                    n = int(nr)
                while os.path.isfile("%s%03d%s" % (fn, n, ext)):
                    n = n + 1
                filename = "%s%03d%s" % (fn, n, ext)

        if openDialog:
            foo = QtGui.QFileDialog.getSaveFileName(
                QtGui.QApplication.activeWindow(), "Output File", filename
            )
            if foo[0]:
                filename = foo[0]
            else:
                filename = None

        return filename

    def resolvePostProcessor(self, job):
        if hasattr(job, "PostProcessor"):
            post = PathPreferences.defaultPostProcessor()
            if job.PostProcessor:
                post = job.PostProcessor
            if post and PostProcessor.exists(post):
                return post
        dlg = DlgSelectPostProcessor()
        return dlg.exec_()

    def GetResources(self):
        return {
            "Pixmap": "Path_Post",
            "MenuText": QT_TRANSLATE_NOOP("Path_Post", "Post Process"),
            "Accel": "P, P",
            "ToolTip": QT_TRANSLATE_NOOP("Path_Post", "Post Process the selected Job"),
        }

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            if FreeCADGui.Selection.getCompleteSelection():
                for o in FreeCAD.ActiveDocument.Objects:
                    if o.Name[:3] == "Job":
                        return True

        return False

    def exportObjectsWith(self, objs, job, needFilename=True):
        PathLog.track()
        # check if the user has a project and has set the default post and
        # output filename
        postArgs = PathPreferences.defaultPostProcessorArgs()
        if hasattr(job, "PostProcessorArgs") and job.PostProcessorArgs:
            postArgs = job.PostProcessorArgs
        elif hasattr(job, "PostProcessor") and job.PostProcessor:
            postArgs = ""

        postname = self.resolvePostProcessor(job)
        filename = "-"
        if postname and needFilename:
            filename = self.resolveFileName(job)

        if postname and filename:
            print("post: %s(%s, %s)" % (postname, filename, postArgs))
            processor = PostProcessor.load(postname)
            gcode = processor.export(objs, filename, postArgs)
            return (False, gcode, filename)
        else:
            return (True, "", filename)

    def buildPostList(self, job):
        """
        Parses the job and returns the list(s) of objects to be written by the post
        Postlist is a list of lists.  Each sublist is intended to be a separate file
        """
        orderby = job.OrderOutputBy

        fixturelist = []
        for f in job.Fixtures:
            # create an object to serve as the fixture path
            fobj = _TempObject()
            fobj.Label = f
            c1 = Path.Command(f)
            c2 = Path.Command(
                "G0 Z"
                + str(
                    job.Stock.Shape.BoundBox.ZMax
                    + job.SetupSheet.ClearanceHeightOffset.Value
                )
            )
            fobj.Path = Path.Path([c1, c2])
            # fobj.InList.append(job)
            fixturelist.append(fobj)

        postlist = []

        if orderby == "Fixture":
            PathLog.debug("Ordering by Fixture")
            # Order by fixture means all operations and tool changes will be completed in one
            # fixture before moving to the next.

            for f in fixturelist:
                scratchpad = [(f, None)]

                # Now generate the gcode
                for obj in job.Operations.Group:
                    if not PathUtil.opProperty(obj, "Active"):
                        continue
                    tc = PathUtil.toolControllerForOp(obj)
                    scratchpad.append((obj, tc))

                sublist = []
                temptool = None
                for item in scratchpad:
                    if item[1] in [temptool, None]:
                        sublist.append(item[0])
                    else:
                        sublist.append(item[1])
                        temptool = item[1]
                        sublist.append(item[0])
                postlist.append(sublist)

        elif orderby == "Tool":
            PathLog.debug("Ordering by Tool")
            # Order by tool means tool changes are minimized.
            # all operations with the current tool are processed in the current
            # fixture before moving to the next fixture.

            currTool = None

            # Now generate the gcode
            curlist = []  # list of ops for tool, will repeat for each fixture
            # sublist = []  # list of ops for output splitting

            for idx, obj in enumerate(job.Operations.Group):

                # check if the operation is active
                active = PathUtil.opProperty(obj, "Active")

                tc = PathUtil.toolControllerForOp(obj)

                if not active:  # pass on any inactive ops
                    continue

                if tc is None:
                    curlist.append((obj, None))
                    continue

                if tc == currTool:
                    curlist.append((obj, tc))
                    continue

                if tc != currTool and currTool is None:  # first TC
                    currTool = tc
                    curlist.append((obj, tc))
                    continue

                if tc != currTool and currTool is not None:  # TC changed
                    if tc.ToolNumber == currTool.ToolNumber:  # Same tool /diff params
                        curlist.append((obj, tc))
                        currTool = tc
                    else:  # Actual Toolchange
                        # dump current state to postlist
                        sublist = []
                        t = None
                        for fixture in fixturelist:
                            sublist.append(fixture)
                            for item in curlist:
                                if item[1] == t:
                                    sublist.append(item[0])
                                else:
                                    sublist.append(item[1])
                                    t = item[1]
                                    sublist.append(item[0])

                        postlist.append(sublist)

                        # set up for next tool group
                        currTool = tc
                        curlist = [(obj, tc)]

            # flush remaining curlist to output
            sublist = []
            t = None
            for fixture in fixturelist:
                sublist.append(fixture)
                for item in curlist:
                    if item[1] == t:
                        sublist.append(item[0])
                    else:
                        sublist.append(item[1])
                        t = item[1]
                        sublist.append(item[0])
            postlist.append(sublist)

        elif orderby == "Operation":
            PathLog.debug("Ordering by Operation")
            # Order by operation means ops are done in each fixture in
            # sequence.
            currTool = None
            # firstFixture = True

            # Now generate the gcode
            for obj in job.Operations.Group:
                scratchpad = []
                tc = PathUtil.toolControllerForOp(obj)
                if not PathUtil.opProperty(obj, "Active"):
                    continue

                PathLog.debug("obj: {}".format(obj.Name))
                for f in fixturelist:

                    scratchpad.append((f, None))
                    scratchpad.append((obj, tc))

                sublist = []
                temptool = None
                for item in scratchpad:
                    if item[1] in [temptool, None]:
                        sublist.append(item[0])
                    else:
                        sublist.append(item[1])
                        temptool = item[1]
                        sublist.append(item[0])
                postlist.append(sublist)

        if job.SplitOutput:
            return postlist
        else:
            finalpostlist = [item for slist in postlist for item in slist]
            return [finalpostlist]

    def Activated(self):
        PathLog.track()
        FreeCAD.ActiveDocument.openTransaction("Post Process the Selected path(s)")
        FreeCADGui.addModule("PathScripts.PathPost")

        # Attempt to figure out what the user wants to post-process
        # If a job is selected, post that.
        # If there's only one job in a document, post it.
        # If a user has selected a subobject of a job, post the job.
        # If multiple jobs and can't guess, ask them.

        selected = FreeCADGui.Selection.getSelectionEx()
        if len(selected) > 1:
            FreeCAD.Console.PrintError(
                "Please select a single job or other path object\n"
            )
            return
        elif len(selected) == 1:
            sel = selected[0].Object
            if sel.Name[:3] == "Job":
                job = sel
            elif hasattr(sel, "Path"):
                try:
                    job = PathUtils.findParentJob(sel)
                except Exception:
                    job = None
            else:
                job = None

        if job is None:
            targetlist = []
            for o in FreeCAD.ActiveDocument.Objects:
                if hasattr(o, "Proxy"):
                    if isinstance(o.Proxy, PathJob.ObjectJob):
                        targetlist.append(o.Label)
            PathLog.debug("Possible post objects: {}".format(targetlist))
            if len(targetlist) > 1:
                jobname, result = QtGui.QInputDialog.getItem(
                    None, translate("Path", "Choose a Path Job"), None, targetlist
                )

                if result is False:
                    return
            else:
                jobname = targetlist[0]
            job = FreeCAD.ActiveDocument.getObject(jobname)

        PathLog.debug("about to postprocess job: {}".format(job.Name))

        postlist = self.buildPostList(job)

        fail = True
        rc = ""
        for slist in postlist:
            (fail, rc, filename) = self.exportObjectsWith(slist, job)
            if fail:
                break

        self.subpart = 1

        if fail:
            FreeCAD.ActiveDocument.abortTransaction()
        else:
            if hasattr(job, "LastPostProcessDate"):
                job.LastPostProcessDate = str(datetime.now())
            if hasattr(job, "LastPostProcessOutput"):
                job.LastPostProcessOutput = filename
            FreeCAD.ActiveDocument.commitTransaction()

        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("Path_Post", CommandPathPost())

FreeCAD.Console.PrintLog("Loading PathPost... done\n")



#   (c) Juergen Riegel (FreeCAD@juergen-riegel.net) 2011      LGPL        *
#                                                                         *
#   This file is part of the FreeCAD CAx development system.              *
#                                                                         *
#   This program is free software; you can redistribute it and/or modify  *
#   it under the terms of the GNU Lesser General Public License (LGPL)    *
#   as published by the Free Software Foundation; either version 2 of     *
#   the License, or (at your option) any later version.                   *
#   for detail see the LICENCE text file.                                 *
#                                                                         *
#   FreeCAD is distributed in the hope that it will be useful,            *
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#   GNU Library General Public License for more details.                  *
#                                                                         *
#   You should have received a copy of the GNU Library General Public     *
#   License along with FreeCAD; if not, write to the Free Software        *
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#   USA                                                                   *
#**************************************************************************

import FreeCAD, unittest, Part
import copy
import math
from FreeCAD import Units
from FreeCAD import Base
App = FreeCAD

from parttests.regression_tests import RegressionTests

#---------------------------------------------------------------------------
# define the test cases to test the FreeCAD Part module
#---------------------------------------------------------------------------


class PartTestCases(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartTest")

    def testBoxCase(self):
        self.Box = self.Doc.addObject("Part::Box","Box")
        self.Doc.recompute()
        self.failUnless(len(self.Box.Shape.Faces)==6)

    def testIssue2985(self):
        v1 = App.Vector(0.0,0.0,0.0)
        v2 = App.Vector(10.0,0.0,0.0)
        v3 = App.Vector(10.0,0.0,10.0)
        v4 = App.Vector(0.0,0.0,10.0)
        edge1 = Part.makeLine(v1, v2)
        edge2 = Part.makeLine(v2, v3)
        edge3 = Part.makeLine(v3, v4)
        edge4 = Part.makeLine(v4, v1)
        # Travis build confirms the crash under macOS
        #result = Part.makeFilledFace([edge1,edge2,edge3,edge4])
        #self.Doc.addObject("Part::Feature","Face").Shape = result
        #self.assertTrue(isinstance(result.Surface, Part.BSplineSurface))

    def tearDown(self):
        #closing doc
        FreeCAD.closeDocument("PartTest")
        #print ("omit closing document for debugging")

class PartTestBSplineCurve(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartTest")

        poles = [[0, 0, 0], [1, 1, 0], [2, 0, 0]]
        self.spline = Part.BSplineCurve()
        self.spline.buildFromPoles(poles)

        poles = [[0, 0, 0], [1, 1, 0], [2, 0, 0], [1, -1, 0]]
        self.nurbs = Part.BSplineCurve()
        self.nurbs.buildFromPolesMultsKnots(poles, (3, 1, 3),(0, 0.5, 1), False, 2)

    def testProperties(self):
        self.assertEqual(self.spline.Continuity, 'CN')
        self.assertEqual(self.spline.Degree, 2)
        self.assertEqual(self.spline.EndPoint, App.Vector(2, 0, 0))
        self.assertEqual(self.spline.FirstParameter, 0.0)
        self.assertEqual(self.spline.FirstUKnotIndex, 1)
        self.assertEqual(self.spline.KnotSequence, [0.0, 0.0, 0.0, 1.0, 1.0, 1.0])
        self.assertEqual(self.spline.LastParameter, 1.0)
        self.assertEqual(self.spline.LastUKnotIndex, 2)
        max_degree = self.spline.MaxDegree
        self.assertEqual(self.spline.NbKnots, 2)
        self.assertEqual(self.spline.NbPoles, 3)
        self.assertEqual(self.spline.StartPoint, App.Vector(0.0, 0.0, 0.0))

    def testGetters(self):
        '''only check if the function doesn't crash'''
        self.spline.getKnot(1)
        self.spline.getKnots()
        self.spline.getMultiplicities()
        self.spline.getMultiplicity(1)
        self.spline.getPole(1)
        self.spline.getPoles()
        self.spline.getPolesAndWeights()
        self.spline.getResolution(0.5)
        self.spline.getWeight(1)
        self.spline.getWeights()

    def testSetters(self):
        spline = copy.copy(self.spline)
        spline.setKnot(1, 0.1)
        spline.setPeriodic()
        spline.setNotPeriodic()
        # spline.setKnots()
        # spline.setOrigin(2)   # not working?
        self.spline.setPole(1, App.Vector([1, 0, 0])) # first parameter 0 gives occ error

    def testIssue2671(self):
        self.Doc = App.newDocument("Issue2671")
        Box = self.Doc.addObject("Part::Box","Box")
        Mirroring = self.Doc.addObject("Part::Mirroring", 'Mirroring')
        Spreadsheet = self.Doc.addObject('Spreadsheet::Sheet', 'Spreadsheet')
        Mirroring.Source = Box
        Mirroring.Base = (8, 5, 25)
        Mirroring.Normal = (0.5, 0.2, 0.9)
        Spreadsheet.set('A1', '=Mirroring.Base.x')
        Spreadsheet.set('B1', '=Mirroring.Base.y')
        Spreadsheet.set('C1', '=Mirroring.Base.z')
        Spreadsheet.set('A2', '=Mirroring.Normal.x')
        Spreadsheet.set('B2', '=Mirroring.Normal.y')
        Spreadsheet.set('C2', '=Mirroring.Normal.z')
        self.Doc.recompute()
        self.assertEqual(Spreadsheet.A1, Units.Quantity('8 mm'))
        self.assertEqual(Spreadsheet.B1, Units.Quantity('5 mm'))
        self.assertEqual(Spreadsheet.C1, Units.Quantity('25 mm'))
        self.assertEqual(Spreadsheet.A2, Units.Quantity('0.5 mm'))
        self.assertEqual(Spreadsheet.B2, Units.Quantity('0.2 mm'))
        self.assertEqual(Spreadsheet.C2, Units.Quantity('0.9 mm'))
        App.closeDocument("Issue2671")

    def testIssue2876(self):
        self.Doc = App.newDocument("Issue2876")
        Cylinder = self.Doc.addObject("Part::Cylinder", "Cylinder")
        Cylinder.Radius = 5
        Pipe = self.Doc.addObject("Part::Thickness", "Pipe")
        Pipe.Faces = (Cylinder, ["Face2", "Face3"])
        Pipe.Mode = 1
        Pipe.Value = -1 # negative wall thickness
        Spreadsheet = self.Doc.addObject('Spreadsheet::Sheet', 'Spreadsheet')
        Spreadsheet.set('A1', 'Pipe OD')
        Spreadsheet.set('B1', 'Pipe WT')
        Spreadsheet.set('C1', 'Pipe ID')
        Spreadsheet.set('A2', '=2*Cylinder.Radius')
        Spreadsheet.set('B2', '=-Pipe.Value')
        Spreadsheet.set('C2', '=2*(Cylinder.Radius + Pipe.Value)')
        self.Doc.recompute()
        self.assertEqual(Spreadsheet.B2, Units.Quantity('1 mm'))
        self.assertEqual(Spreadsheet.C2, Units.Quantity('8 mm'))
        App.closeDocument("Issue2876")

    def testSubElements(self):
        box = Part.makeBox(1, 1, 1)
        with self.assertRaises(ValueError):
            box.getElement("InvalidName")
        with self.assertRaises(ValueError):
            box.getElement("Face6_abc")
        with self.assertRaises(Part.OCCError):
            box.getElement("Face7")

    def tearDown(self):
        #closing doc
        FreeCAD.closeDocument("PartTest")

class PartTestNormals(unittest.TestCase):
    def setUp(self):
        self.face = Part.makePlane(1, 1)

    def testFaceNormal(self):
        self.assertEqual(self.face.normalAt(0, 0), Base.Vector(0, 0, 1))
        self.assertEqual(self.face.Surface.normal(0, 0), Base.Vector(0, 0, 1))

    def testReverseOrientation(self):
        self.face.reverse()
        self.assertEqual(self.face.normalAt(0, 0), Base.Vector(0, 0, -1))
        self.assertEqual(self.face.Surface.normal(0, 0), Base.Vector(0, 0, 1))

    def testPlacement(self):
        self.face.reverse()
        self.face.Placement.Rotation.Angle = 1
        self.face.Placement.Rotation.Axis = (1,1,1)
        vec = Base.Vector(-0.63905, 0.33259, -0.69353)
        self.assertGreater(self.face.normalAt(0, 0).dot(vec), 0.9999)
        self.assertLess(self.face.Surface.normal(0, 0).dot(vec), -0.9999)

    def tearDown(self):
        pass

class PartTestCircle2D(unittest.TestCase):
    def testValidCircle(self):
        p1 = App.Base.Vector2d(0.01, 0.01)
        p2 = App.Base.Vector2d(0.02, 0.02)
        p3 = App.Base.Vector2d(0.01, -0.01)
        Part.Geom2d.Circle2d.getCircleCenter(p1, p2, p3)

    def testCollinearPoints(self):
        p1 = App.Base.Vector2d(0.01, 0.01)
        p2 = App.Base.Vector2d(0.02, 0.02)
        p3 = App.Base.Vector2d(0.04, 0.0399)
        with self.assertRaises(ValueError):
            Part.Geom2d.Circle2d.getCircleCenter(p1, p2, p3)

class PartTestCone(unittest.TestCase):
    def testderivatives(self):
        def get_dn(surface, u, v):
            pos = surface.value(u, v)
            v10 = surface.getDN(u, v, 1, 0)
            v01 = surface.getDN(u, v, 0, 1)
            v11 = surface.getDN(u, v, 1, 1)
            return (pos, v10, v01, v11)

        cone = Part.Cone()
        cone.SemiAngle = 0.2
        cone.Radius = 2.0

        u, v = (5.0, 5.0)
        vp, v1, v2, v3 = get_dn(cone, u, v)

        shape = cone.toShape(0, 2*math.pi, 0, 10)
        shape = shape.toNurbs()
        spline = shape.Face1.Surface

        u, v = spline.parameter(vp)
        wp, w1, w2, w3 = get_dn(spline, u, v)

        self.assertAlmostEqual(vp.distanceToPoint(wp), 0)
        self.assertAlmostEqual(v1.getAngle(w1), 0)
        self.assertAlmostEqual(v2.getAngle(w2), 0)
        self.assertAlmostEqual(v3.getAngle(w3), 0)

class PartTestChFi2dAlgos(unittest.TestCase):
    def testChFi2d_FilletAlgo(self):
        v = FreeCAD.Vector
        edge1 = Part.makeLine(v(0,0,0), v(0,10,0))
        edge2 = Part.makeLine(v(0,10,0), v(10,10,0))
        wire = Part.Wire([edge1, edge2])
        pln = Part.Plane()

        with self.assertRaises(TypeError):
            alg = Part.ChFi2d.FilletAlgo(pln)

        alg = Part.ChFi2d.FilletAlgo()
        with self.assertRaises(TypeError):
            alg.init()

        print (alg)
        # Test without shape
        with self.assertRaises(Base.CADKernelError):
            alg.perform(1)

        with self.assertRaises(TypeError):
            alg.perform()

        alg = Part.ChFi2d.FilletAlgo(wire, pln)
        alg.init(edge1, edge2, pln)
        alg.init(wire, pln)

        alg = Part.ChFi2d.FilletAlgo(edge1, edge2, pln)
        alg.perform(1.0)

        with self.assertRaises(TypeError):
            alg.numberOfResults()

        with self.assertRaises(TypeError):
            alg.result(1)

        self.assertEqual(alg.numberOfResults(Base.Vector(0,10,0)), 1)
        result = alg.result(Base.Vector(0,10,0))
        curve = result[0].Curve
        self.assertEqual(type(curve), Part.Circle)
        self.assertEqual(curve.Axis, pln.Axis)
        self.assertEqual(curve.Radius, 1.0)

    def testChFi2d_AnaFilletAlgo(self):
        v = FreeCAD.Vector
        edge1 = Part.makeLine(v(0,0,0), v(0,10,0))
        edge2 = Part.makeLine(v(0,10,0), v(10,10,0))
        wire = Part.Wire([edge1, edge2])
        pln = Part.Plane()

        with self.assertRaises(TypeError):
            alg = Part.ChFi2d.AnaFilletAlgo(pln)

        alg = Part.ChFi2d.AnaFilletAlgo()
        with self.assertRaises(TypeError):
            alg.init()

        print (alg)
        # Test without shape
        self.assertFalse(alg.perform(1))

        with self.assertRaises(TypeError):
            alg.perform()

        alg = Part.ChFi2d.AnaFilletAlgo(wire, pln)
        alg.init(edge1, edge2, pln)
        alg.init(wire, pln)

        alg = Part.ChFi2d.AnaFilletAlgo(edge1, edge2, pln)
        alg.perform(1.0)

        with self.assertRaises(TypeError):
            alg.result(1)

        result = alg.result()
        curve = result[0].Curve
        self.assertEqual(type(curve), Part.Circle)
        self.assertEqual(curve.Radius, 1.0)

    def testChFi2d_ChamferAPI(self):
        v = FreeCAD.Vector
        edge1 = Part.makeLine(v(0,0,0), v(0,10,0))
        edge2 = Part.makeLine(v(0,10,0), v(10,10,0))
        wire = Part.Wire([edge1, edge2])

        with self.assertRaises(TypeError):
            alg = Part.ChFi2d.ChamferAPI(edge1)

        alg = Part.ChFi2d.ChamferAPI(wire)
        with self.assertRaises(TypeError):
            alg.init()

        print (alg)

        with self.assertRaises(TypeError):
            alg.perform(1)

        alg = Part.ChFi2d.ChamferAPI(wire)
        alg.init(edge1, edge2)
        alg.init(wire)

        alg = Part.ChFi2d.ChamferAPI(edge1, edge2)
        alg.perform()

        with self.assertRaises(TypeError):
            alg.result(1)

        result = alg.result(1.0, 1.0)
        curve = result[0].Curve
        self.assertEqual(type(curve), Part.Line)



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2018 Kresimir Tusek <kresimir.tusek@gmail.com>          *
# *   Copyright (c) 2019-2021 Schildkroet                                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************

import PathScripts.PathOp as PathOp
import PathScripts.PathUtils as PathUtils
import PathScripts.PathLog as PathLog
import PathScripts.PathGeom as PathGeom
import Path
import FreeCAD
import time
import json
import math
import area
from PySide.QtCore import QT_TRANSLATE_NOOP

if FreeCAD.GuiUp:
    from pivy import coin
    import FreeCADGui

__doc__ = "Class and implementation of the Adaptive path operation."

# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

Part = LazyLoader("Part", globals(), "Part")
# TechDraw = LazyLoader('TechDraw', globals(), 'TechDraw')
FeatureExtensions = LazyLoader(
    "PathScripts.PathFeatureExtensions", globals(), "PathScripts.PathFeatureExtensions"
)
DraftGeomUtils = LazyLoader("DraftGeomUtils", globals(), "DraftGeomUtils")


if False:
    PathLog.setLevel(PathLog.Level.DEBUG, PathLog.thisModule())
    PathLog.trackModule(PathLog.thisModule())
else:
    PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())


translate = FreeCAD.Qt.translate


def convertTo2d(pathArray):
    output = []
    for path in pathArray:
        pth2 = []
        for edge in path:
            for pt in edge:
                pth2.append([pt[0], pt[1]])
        output.append(pth2)
    return output


sceneGraph = None
scenePathNodes = []  # for scene cleanup aftewards
topZ = 10


def sceneDrawPath(path, color=(0, 0, 1)):
    coPoint = coin.SoCoordinate3()

    pts = []
    for pt in path:
        pts.append([pt[0], pt[1], topZ])

    coPoint.point.setValues(0, len(pts), pts)
    ma = coin.SoBaseColor()
    ma.rgb = color
    li = coin.SoLineSet()
    li.numVertices.setValue(len(pts))
    pathNode = coin.SoSeparator()
    pathNode.addChild(coPoint)
    pathNode.addChild(ma)
    pathNode.addChild(li)
    sceneGraph.addChild(pathNode)
    scenePathNodes.append(pathNode)  # for scene cleanup afterwards


def sceneClean():
    for n in scenePathNodes:
        sceneGraph.removeChild(n)

    del scenePathNodes[:]


def discretize(edge, flipDirection=False):
    pts = edge.discretize(Deflection=0.002)
    if flipDirection:
        pts.reverse()

    return pts


def CalcHelixConePoint(height, cur_z, radius, angle):
    x = ((height - cur_z) / height) * radius * math.cos(math.radians(angle) * cur_z)
    y = ((height - cur_z) / height) * radius * math.sin(math.radians(angle) * cur_z)
    z = cur_z

    return {"X": x, "Y": y, "Z": z}


def GenerateGCode(op, obj, adaptiveResults, helixDiameter):
    if len(adaptiveResults) == 0 or len(adaptiveResults[0]["AdaptivePaths"]) == 0:
        return

    # minLiftDistance = op.tool.Diameter
    helixRadius = 0
    for region in adaptiveResults:
        p1 = region["HelixCenterPoint"]
        p2 = region["StartPoint"]
        r = math.sqrt(
            (p1[0] - p2[0]) * (p1[0] - p2[0]) + (p1[1] - p2[1]) * (p1[1] - p2[1])
        )
        if r > helixRadius:
            helixRadius = r

    stepDown = obj.StepDown.Value
    passStartDepth = obj.StartDepth.Value

    if stepDown < 0.1:
        stepDown = 0.1

    length = 2 * math.pi * helixRadius

    if float(obj.HelixAngle) < 1:
        obj.HelixAngle = 1
    if float(obj.HelixAngle) > 89:
        obj.HelixAngle = 89

    if float(obj.HelixConeAngle) < 0:
        obj.HelixConeAngle = 0

    helixAngleRad = math.pi * float(obj.HelixAngle) / 180.0
    depthPerOneCircle = length * math.tan(helixAngleRad)
    # print("Helix circle depth: {}".format(depthPerOneCircle))

    stepUp = obj.LiftDistance.Value
    if stepUp < 0:
        stepUp = 0

    finish_step = obj.FinishDepth.Value if hasattr(obj, "FinishDepth") else 0.0
    if finish_step > stepDown:
        finish_step = stepDown

    depth_params = PathUtils.depth_params(
        clearance_height=obj.ClearanceHeight.Value,
        safe_height=obj.SafeHeight.Value,
        start_depth=obj.StartDepth.Value,
        step_down=stepDown,
        z_finish_step=finish_step,
        final_depth=obj.FinalDepth.Value,
        user_depths=None,
    )

    # ml: this is dangerous because it'll hide all unused variables hence forward
    #     however, I don't know what lx and ly signify so I'll leave them for now
    # lx = adaptiveResults[0]["HelixCenterPoint"][0]
    # ly = adaptiveResults[0]["HelixCenterPoint"][1]
    lz = passStartDepth
    step = 0

    for passEndDepth in depth_params.data:
        step = step + 1

        for region in adaptiveResults:
            startAngle = math.atan2(
                region["StartPoint"][1] - region["HelixCenterPoint"][1],
                region["StartPoint"][0] - region["HelixCenterPoint"][0],
            )

            # lx = region["HelixCenterPoint"][0]
            # ly = region["HelixCenterPoint"][1]

            passDepth = passStartDepth - passEndDepth

            p1 = region["HelixCenterPoint"]
            p2 = region["StartPoint"]
            helixRadius = math.sqrt(
                (p1[0] - p2[0]) * (p1[0] - p2[0]) + (p1[1] - p2[1]) * (p1[1] - p2[1])
            )

            # Helix ramp
            if helixRadius > 0.01:
                r = helixRadius - 0.01

                maxfi = passDepth / depthPerOneCircle * 2 * math.pi
                fi = 0
                offsetFi = -maxfi + startAngle - math.pi / 16

                helixStart = [
                    region["HelixCenterPoint"][0] + r * math.cos(offsetFi),
                    region["HelixCenterPoint"][1] + r * math.sin(offsetFi),
                ]

                op.commandlist.append(
                    Path.Command("(Helix to depth: %f)" % passEndDepth)
                )

                if obj.UseHelixArcs is False:
                    # rapid move to start point
                    op.commandlist.append(
                        Path.Command("G0", {"Z": obj.ClearanceHeight.Value})
                    )
                    op.commandlist.append(
                        Path.Command(
                            "G0",
                            {
                                "X": helixStart[0],
                                "Y": helixStart[1],
                                "Z": obj.ClearanceHeight.Value,
                            },
                        )
                    )

                    # rapid move to safe height
                    op.commandlist.append(
                        Path.Command(
                            "G0",
                            {
                                "X": helixStart[0],
                                "Y": helixStart[1],
                                "Z": obj.SafeHeight.Value,
                            },
                        )
                    )

                    # move to start depth
                    op.commandlist.append(
                        Path.Command(
                            "G1",
                            {
                                "X": helixStart[0],
                                "Y": helixStart[1],
                                "Z": passStartDepth,
                                "F": op.vertFeed,
                            },
                        )
                    )

                    if obj.HelixConeAngle == 0:
                        while fi < maxfi:
                            x = region["HelixCenterPoint"][0] + r * math.cos(
                                fi + offsetFi
                            )
                            y = region["HelixCenterPoint"][1] + r * math.sin(
                                fi + offsetFi
                            )
                            z = passStartDepth - fi / maxfi * (
                                passStartDepth - passEndDepth
                            )
                            op.commandlist.append(
                                Path.Command(
                                    "G1", {"X": x, "Y": y, "Z": z, "F": op.vertFeed}
                                )
                            )
                            # lx = x
                            # ly = y
                            fi = fi + math.pi / 16

                        # one more circle at target depth to make sure center is cleared
                        maxfi = maxfi + 2 * math.pi
                        while fi < maxfi:
                            x = region["HelixCenterPoint"][0] + r * math.cos(
                                fi + offsetFi
                            )
                            y = region["HelixCenterPoint"][1] + r * math.sin(
                                fi + offsetFi
                            )
                            z = passEndDepth
                            op.commandlist.append(
                                Path.Command(
                                    "G1", {"X": x, "Y": y, "Z": z, "F": op.horizFeed}
                                )
                            )
                            # lx = x
                            # ly = y
                            fi = fi + math.pi / 16

                    else:
                        # Cone
                        _HelixAngle = 360 - (float(obj.HelixAngle) * 4)

                        if obj.HelixConeAngle > 6:
                            obj.HelixConeAngle = 6

                        helixRadius *= 0.9

                        # Calculate everything
                        helix_height = passStartDepth - passEndDepth
                        r_extra = helix_height * math.tan(
                            math.radians(obj.HelixConeAngle)
                        )
                        HelixTopRadius = helixRadius + r_extra
                        helix_full_height = HelixTopRadius * (
                            math.cos(math.radians(obj.HelixConeAngle))
                            / math.sin(math.radians(obj.HelixConeAngle))
                        )

                        # Start height
                        z = passStartDepth
                        i = 0

                        # Default step down
                        z_step = 0.05

                        # Bigger angle, smaller step down
                        if _HelixAngle > 120:
                            z_step = 0.025
                        if _HelixAngle > 240:
                            z_step = 0.015

                        p = None
                        # Calculate conical helix
                        while z >= passEndDepth:
                            if z < passEndDepth:
                                z = passEndDepth

                            p = CalcHelixConePoint(
                                helix_full_height, i, HelixTopRadius, _HelixAngle
                            )
                            op.commandlist.append(
                                Path.Command(
                                    "G1",
                                    {
                                        "X": p["X"] + region["HelixCenterPoint"][0],
                                        "Y": p["Y"] + region["HelixCenterPoint"][1],
                                        "Z": z,
                                        "F": op.vertFeed,
                                    },
                                )
                            )
                            z = z - z_step
                            i = i + z_step

                        # Calculate some stuff for arcs at bottom
                        p["X"] = p["X"] + region["HelixCenterPoint"][0]
                        p["Y"] = p["Y"] + region["HelixCenterPoint"][1]
                        x_m = (
                            region["HelixCenterPoint"][0]
                            - p["X"]
                            + region["HelixCenterPoint"][0]
                        )
                        y_m = (
                            region["HelixCenterPoint"][1]
                            - p["Y"]
                            + region["HelixCenterPoint"][1]
                        )
                        i_off = (x_m - p["X"]) / 2
                        j_off = (y_m - p["Y"]) / 2

                        # One more circle at target depth to make sure center is cleared
                        op.commandlist.append(
                            Path.Command(
                                "G3",
                                {
                                    "X": x_m,
                                    "Y": y_m,
                                    "Z": passEndDepth,
                                    "I": i_off,
                                    "J": j_off,
                                    "F": op.horizFeed,
                                },
                            )
                        )
                        op.commandlist.append(
                            Path.Command(
                                "G3",
                                {
                                    "X": p["X"],
                                    "Y": p["Y"],
                                    "Z": passEndDepth,
                                    "I": -i_off,
                                    "J": -j_off,
                                    "F": op.horizFeed,
                                },
                            )
                        )

                else:
                    # Use arcs for helix - no conical shape support
                    helixStart = [
                        region["HelixCenterPoint"][0] + r,
                        region["HelixCenterPoint"][1],
                    ]

                    # rapid move to start point
                    op.commandlist.append(
                        Path.Command("G0", {"Z": obj.ClearanceHeight.Value})
                    )
                    op.commandlist.append(
                        Path.Command(
                            "G0",
                            {
                                "X": helixStart[0],
                                "Y": helixStart[1],
                                "Z": obj.ClearanceHeight.Value,
                            },
                        )
                    )

                    # rapid move to safe height
                    op.commandlist.append(
                        Path.Command(
                            "G0",
                            {
                                "X": helixStart[0],
                                "Y": helixStart[1],
                                "Z": obj.SafeHeight.Value,
                            },
                        )
                    )

                    # move to start depth
                    op.commandlist.append(
                        Path.Command(
                            "G1",
                            {
                                "X": helixStart[0],
                                "Y": helixStart[1],
                                "Z": passStartDepth,
                                "F": op.vertFeed,
                            },
                        )
                    )

                    x = region["HelixCenterPoint"][0] + r
                    y = region["HelixCenterPoint"][1]

                    curDep = passStartDepth
                    while curDep > (passEndDepth + depthPerOneCircle):
                        op.commandlist.append(
                            Path.Command(
                                "G2",
                                {
                                    "X": x - (2 * r),
                                    "Y": y,
                                    "Z": curDep - (depthPerOneCircle / 2),
                                    "I": -r,
                                    "F": op.vertFeed,
                                },
                            )
                        )
                        op.commandlist.append(
                            Path.Command(
                                "G2",
                                {
                                    "X": x,
                                    "Y": y,
                                    "Z": curDep - depthPerOneCircle,
                                    "I": r,
                                    "F": op.vertFeed,
                                },
                            )
                        )
                        curDep = curDep - depthPerOneCircle

                    lastStep = curDep - passEndDepth
                    if lastStep > (depthPerOneCircle / 2):
                        op.commandlist.append(
                            Path.Command(
                                "G2",
                                {
                                    "X": x - (2 * r),
                                    "Y": y,
                                    "Z": curDep - (lastStep / 2),
                                    "I": -r,
                                    "F": op.vertFeed,
                                },
                            )
                        )
                        op.commandlist.append(
                            Path.Command(
                                "G2",
                                {
                                    "X": x,
                                    "Y": y,
                                    "Z": passEndDepth,
                                    "I": r,
                                    "F": op.vertFeed,
                                },
                            )
                        )
                    else:
                        op.commandlist.append(
                            Path.Command(
                                "G2",
                                {
                                    "X": x - (2 * r),
                                    "Y": y,
                                    "Z": passEndDepth,
                                    "I": -r,
                                    "F": op.vertFeed,
                                },
                            )
                        )
                        op.commandlist.append(
                            Path.Command(
                                "G1",
                                {"X": x, "Y": y, "Z": passEndDepth, "F": op.vertFeed},
                            )
                        )

                    # one more circle at target depth to make sure center is cleared
                    op.commandlist.append(
                        Path.Command(
                            "G2",
                            {
                                "X": x - (2 * r),
                                "Y": y,
                                "Z": passEndDepth,
                                "I": -r,
                                "F": op.horizFeed,
                            },
                        )
                    )
                    op.commandlist.append(
                        Path.Command(
                            "G2",
                            {
                                "X": x,
                                "Y": y,
                                "Z": passEndDepth,
                                "I": r,
                                "F": op.horizFeed,
                            },
                        )
                    )
                    # lx = x
                    # ly = y

            else:  # no helix entry
                # rapid move to clearance height
                op.commandlist.append(
                    Path.Command("G0", {"Z": obj.ClearanceHeight.Value})
                )
                op.commandlist.append(
                    Path.Command(
                        "G0",
                        {
                            "X": region["StartPoint"][0],
                            "Y": region["StartPoint"][1],
                            "Z": obj.ClearanceHeight.Value,
                        },
                    )
                )
                # straight plunge to target depth
                op.commandlist.append(
                    Path.Command(
                        "G1",
                        {
                            "X": region["StartPoint"][0],
                            "Y": region["StartPoint"][1],
                            "Z": passEndDepth,
                            "F": op.vertFeed,
                        },
                    )
                )

            lz = passEndDepth
            z = obj.ClearanceHeight.Value
            op.commandlist.append(Path.Command("(Adaptive - depth: %f)" % passEndDepth))

            # add adaptive paths
            for pth in region["AdaptivePaths"]:
                motionType = pth[0]  # [0] contains motion type

                for pt in pth[1]:  # [1] contains list of points
                    x = pt[0]
                    y = pt[1]

                    # dist = math.sqrt((x-lx)*(x-lx) + (y-ly)*(y-ly))

                    if motionType == area.AdaptiveMotionType.Cutting:
                        z = passEndDepth
                        if z != lz:
                            op.commandlist.append(
                                Path.Command("G1", {"Z": z, "F": op.vertFeed})
                            )

                        op.commandlist.append(
                            Path.Command("G1", {"X": x, "Y": y, "F": op.horizFeed})
                        )

                    elif motionType == area.AdaptiveMotionType.LinkClear:
                        z = passEndDepth + stepUp
                        if z != lz:
                            op.commandlist.append(Path.Command("G0", {"Z": z}))

                        op.commandlist.append(Path.Command("G0", {"X": x, "Y": y}))

                    elif motionType == area.AdaptiveMotionType.LinkNotClear:
                        z = obj.ClearanceHeight.Value
                        if z != lz:
                            op.commandlist.append(Path.Command("G0", {"Z": z}))

                        op.commandlist.append(Path.Command("G0", {"X": x, "Y": y}))

                    # elif motionType == area.AdaptiveMotionType.LinkClearAtPrevPass:
                    #     if lx!=x or ly!=y:
                    #         op.commandlist.append(Path.Command("G0", { "X": lx, "Y":ly, "Z":passStartDepth+stepUp}))
                    #     op.commandlist.append(Path.Command("G0", { "X": x, "Y":y, "Z":passStartDepth+stepUp}))

                    # lx = x
                    # ly = y
                    lz = z

            # return to safe height in this Z pass
            z = obj.ClearanceHeight.Value
            if z != lz:
                op.commandlist.append(Path.Command("G0", {"Z": z}))

            lz = z

        passStartDepth = passEndDepth

        # return to safe height in this Z pass
        z = obj.ClearanceHeight.Value
        if z != lz:
            op.commandlist.append(Path.Command("G0", {"Z": z}))

        lz = z

    z = obj.ClearanceHeight.Value
    if z != lz:
        op.commandlist.append(Path.Command("G0", {"Z": z}))


def Execute(op, obj):
    global sceneGraph
    global topZ

    if FreeCAD.GuiUp:
        sceneGraph = FreeCADGui.ActiveDocument.ActiveView.getSceneGraph()

    PathLog.info("*** Adaptive toolpath processing started...\n")

    # hide old toolpaths during recalculation
    obj.Path = Path.Path("(Calculating...)")

    if FreeCAD.GuiUp:
        # store old visibility state
        job = op.getJob(obj)
        oldObjVisibility = obj.ViewObject.Visibility
        oldJobVisibility = job.ViewObject.Visibility

        obj.ViewObject.Visibility = False
        job.ViewObject.Visibility = False

        FreeCADGui.updateGui()

    try:
        helixDiameter = obj.HelixDiameterLimit.Value
        topZ = op.stock.Shape.BoundBox.ZMax
        obj.Stopped = False
        obj.StopProcessing = False
        if obj.Tolerance < 0.001:
            obj.Tolerance = 0.001

        # Get list of working edges for adaptive algorithm
        pathArray = op.pathArray
        if not pathArray:
            PathLog.error("No wire data returned.")
            return

        path2d = convertTo2d(pathArray)

        stockPaths = []
        if hasattr(op.stock, "StockType") and op.stock.StockType == "CreateCylinder":
            stockPaths.append([discretize(op.stock.Shape.Edges[0])])

        else:
            stockBB = op.stock.Shape.BoundBox
            v = []
            v.append(FreeCAD.Vector(stockBB.XMin, stockBB.YMin, 0))
            v.append(FreeCAD.Vector(stockBB.XMax, stockBB.YMin, 0))
            v.append(FreeCAD.Vector(stockBB.XMax, stockBB.YMax, 0))
            v.append(FreeCAD.Vector(stockBB.XMin, stockBB.YMax, 0))
            v.append(FreeCAD.Vector(stockBB.XMin, stockBB.YMin, 0))
            stockPaths.append([v])

        stockPath2d = convertTo2d(stockPaths)

        # opType = area.AdaptiveOperationType.ClearingInside  # Commented out per LGTM suggestion
        if obj.OperationType == "Clearing":
            if obj.Side == "Outside":
                opType = area.AdaptiveOperationType.ClearingOutside

            else:
                opType = area.AdaptiveOperationType.ClearingInside

        else:  # profiling
            if obj.Side == "Outside":
                opType = area.AdaptiveOperationType.ProfilingOutside

            else:
                opType = area.AdaptiveOperationType.ProfilingInside

        keepToolDownRatio = 3.0
        if hasattr(obj, "KeepToolDownRatio"):
            keepToolDownRatio = float(obj.KeepToolDownRatio)

        # put here all properties that influence calculation of adaptive base paths,

        inputStateObject = {
            "tool": float(op.tool.Diameter),
            "tolerance": float(obj.Tolerance),
            "geometry": path2d,
            "stockGeometry": stockPath2d,
            "stepover": float(obj.StepOver),
            "effectiveHelixDiameter": float(helixDiameter),
            "operationType": obj.OperationType,
            "side": obj.Side,
            "forceInsideOut": obj.ForceInsideOut,
            "finishingProfile": obj.FinishingProfile,
            "keepToolDownRatio": keepToolDownRatio,
            "stockToLeave": float(obj.StockToLeave),
        }

        inputStateChanged = False
        adaptiveResults = None

        if obj.AdaptiveOutputState is not None and obj.AdaptiveOutputState != "":
            adaptiveResults = obj.AdaptiveOutputState

        if json.dumps(obj.AdaptiveInputState) != json.dumps(inputStateObject):
            inputStateChanged = True
            adaptiveResults = None

        # progress callback fn, if return true it will stop processing
        def progressFn(tpaths):
            if FreeCAD.GuiUp:
                for (
                    path
                ) in (
                    tpaths
                ):  # path[0] contains the MotionType, #path[1] contains list of points
                    if path[0] == area.AdaptiveMotionType.Cutting:
                        sceneDrawPath(path[1], (0, 0, 1))

                    else:
                        sceneDrawPath(path[1], (1, 0, 1))

                FreeCADGui.updateGui()

            return obj.StopProcessing

        start = time.time()

        if inputStateChanged or adaptiveResults is None:
            a2d = area.Adaptive2d()
            a2d.stepOverFactor = 0.01 * obj.StepOver
            a2d.toolDiameter = float(op.tool.Diameter)
            a2d.helixRampDiameter = helixDiameter
            a2d.keepToolDownDistRatio = keepToolDownRatio
            a2d.stockToLeave = float(obj.StockToLeave)
            a2d.tolerance = float(obj.Tolerance)
            a2d.forceInsideOut = obj.ForceInsideOut
            a2d.finishingProfile = obj.FinishingProfile
            a2d.opType = opType

            # EXECUTE
            results = a2d.Execute(stockPath2d, path2d, progressFn)

            # need to convert results to python object to be JSON serializable
            adaptiveResults = []
            for result in results:
                adaptiveResults.append(
                    {
                        "HelixCenterPoint": result.HelixCenterPoint,
                        "StartPoint": result.StartPoint,
                        "AdaptivePaths": result.AdaptivePaths,
                        "ReturnMotionType": result.ReturnMotionType,
                    }
                )

        # GENERATE
        GenerateGCode(op, obj, adaptiveResults, helixDiameter)

        if not obj.StopProcessing:
            PathLog.info("*** Done. Elapsed time: %f sec\n\n" % (time.time() - start))
            obj.AdaptiveOutputState = adaptiveResults
            obj.AdaptiveInputState = inputStateObject

        else:
            PathLog.info(
                "*** Processing cancelled (after: %f sec).\n\n" % (time.time() - start)
            )

    finally:
        if FreeCAD.GuiUp:
            obj.ViewObject.Visibility = oldObjVisibility
            job.ViewObject.Visibility = oldJobVisibility
            sceneClean()


def _get_working_edges(op, obj):
    """_get_working_edges(op, obj)...
    Compile all working edges from the Base Geometry selection (obj.Base)
    for the current operation.
    Additional modifications to selected region(face), such as extensions,
    should be placed within this function.
    """
    all_regions = list()
    edge_list = list()
    avoidFeatures = list()
    rawEdges = list()

    # Get extensions and identify faces to avoid
    extensions = FeatureExtensions.getExtensions(obj)
    for e in extensions:
        if e.avoid:
            avoidFeatures.append(e.feature)

    # Get faces selected by user
    for base, subs in obj.Base:
        for sub in subs:
            if sub.startswith("Face"):
                if sub not in avoidFeatures:
                    if obj.UseOutline:
                        face = base.Shape.getElement(sub)
                        # get outline with wire_A method used in PocketShape, but it does not play nicely later
                        # wire_A = TechDraw.findShapeOutline(face, 1, FreeCAD.Vector(0.0, 0.0, 1.0))
                        wire_B = face.Wires[0]
                        shape = Part.Face(wire_B)
                    else:
                        shape = base.Shape.getElement(sub)
                    all_regions.append(shape)
            elif sub.startswith("Edge"):
                # Save edges for later processing
                rawEdges.append(base.Shape.getElement(sub))
    # Efor

    # Process selected edges
    if rawEdges:
        edgeWires = DraftGeomUtils.findWires(rawEdges)
        if edgeWires:
            for w in edgeWires:
                for e in w.Edges:
                    edge_list.append([discretize(e)])

    # Apply regular Extensions
    op.exts = []
    for ext in extensions:
        if not ext.avoid:
            wire = ext.getWire()
            if wire:
                for f in ext.getExtensionFaces(wire):
                    op.exts.append(f)
                    all_regions.append(f)

    # Second face-combining method attempted
    horizontal = PathGeom.combineHorizontalFaces(all_regions)
    if horizontal:
        obj.removalshape = Part.makeCompound(horizontal)
        for f in horizontal:
            for w in f.Wires:
                for e in w.Edges:
                    edge_list.append([discretize(e)])

    return edge_list


class PathAdaptive(PathOp.ObjectOp):
    def opFeatures(self, obj):
        """opFeatures(obj) ... returns the OR'ed list of features used and supported by the operation.
        The default implementation returns "FeatureTool | FeatureDepths | FeatureHeights | FeatureStartPoint"
        Should be overwritten by subclasses."""
        return (
            PathOp.FeatureTool
            | PathOp.FeatureBaseEdges
            | PathOp.FeatureDepths
            | PathOp.FeatureFinishDepth
            | PathOp.FeatureStepDown
            | PathOp.FeatureHeights
            | PathOp.FeatureBaseGeometry
            | PathOp.FeatureCoolant
            | PathOp.FeatureLocations
        )

    @classmethod
    def propertyEnumerations(self, dataType="data"):
        """helixOpPropertyEnumerations(dataType="data")... return property enumeration lists of specified dataType.
        Args:
            dataType = 'data', 'raw', 'translated'
        Notes:
        'data' is list of internal string literals used in code
        'raw' is list of (translated_text, data_string) tuples
        'translated' is list of translated string literals
        """

        # Enumeration lists for App::PropertyEnumeration properties
        enums = {
            "Side": [
                (translate("Path_Adaptive", "Outside"), "Outside"),
                (translate("Path_Adaptive", "Inside"), "Inside"),
            ],  # this is the direction that the profile runs
            "OperationType": [
                (translate("Path_Adaptive", "Clearing"), "Clearing"),
                (translate("Path_Adaptive", "Profiling"), "Profiling"),
            ],  # side of profile that cutter is on in relation to direction of profile
        }

        if dataType == "raw":
            return enums

        data = list()
        idx = 0 if dataType == "translated" else 1

        PathLog.debug(enums)

        for k, v in enumerate(enums):
            data.append((v, [tup[idx] for tup in enums[v]]))
        PathLog.debug(data)

        return data

    def initOperation(self, obj):
        """initOperation(obj) ... implement to create additional properties.
        Should be overwritten by subclasses."""
        obj.addProperty(
            "App::PropertyEnumeration",
            "Side",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Side of selected faces that tool should cut",
            ),
        )
        # obj.Side = [
        #     "Outside",
        #     "Inside",
        # ]  # side of profile that cutter is on in relation to direction of profile

        obj.addProperty(
            "App::PropertyEnumeration",
            "OperationType",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Type of adaptive operation",
            ),
        )
        # obj.OperationType = [
        #     "Clearing",
        #     "Profiling",
        # ]  # side of profile that cutter is on in relation to direction of profile

        obj.addProperty(
            "App::PropertyFloat",
            "Tolerance",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Influences accuracy and performance",
            ),
        )
        obj.addProperty(
            "App::PropertyPercent",
            "StepOver",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Percent of cutter diameter to step over on each pass",
            ),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "LiftDistance",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Lift distance for rapid moves",
            ),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "KeepToolDownRatio",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Max length of keep tool down path compared to direct distance between points",
            ),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "StockToLeave",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "How much stock to leave (i.e. for finishing operation)",
            ),
        )
        obj.addProperty(
            "App::PropertyBool",
            "ForceInsideOut",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Force plunging into material inside and clearing towards the edges",
            ),
        )
        obj.addProperty(
            "App::PropertyBool",
            "FinishingProfile",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "To take a finishing profile path at the end",
            ),
        )
        obj.addProperty(
            "App::PropertyBool",
            "Stopped",
            "Adaptive",
            QT_TRANSLATE_NOOP("App::Property", "Stop processing"),
        )
        obj.setEditorMode("Stopped", 2)  # hide this property

        obj.addProperty(
            "App::PropertyBool",
            "StopProcessing",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Stop processing",
            ),
        )
        obj.setEditorMode("StopProcessing", 2)  # hide this property

        obj.addProperty(
            "App::PropertyBool",
            "UseHelixArcs",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Use Arcs (G2) for helix ramp",
            ),
        )

        obj.addProperty(
            "App::PropertyPythonObject",
            "AdaptiveInputState",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Internal input state",
            ),
        )
        obj.addProperty(
            "App::PropertyPythonObject",
            "AdaptiveOutputState",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Internal output state",
            ),
        )
        obj.setEditorMode("AdaptiveInputState", 2)  # hide this property
        obj.setEditorMode("AdaptiveOutputState", 2)  # hide this property
        obj.addProperty(
            "App::PropertyAngle",
            "HelixAngle",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Helix ramp entry angle (degrees)",
            ),
        )
        obj.addProperty(
            "App::PropertyAngle",
            "HelixConeAngle",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Helix cone angle (degrees)",
            ),
        )
        obj.addProperty(
            "App::PropertyLength",
            "HelixDiameterLimit",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Limit helix entry diameter, if limit larger than tool diameter or 0, tool diameter is used",
            ),
        )

        obj.addProperty(
            "App::PropertyBool",
            "UseOutline",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Uses the outline of the base geometry.",
            ),
        )

        obj.addProperty(
            "Part::PropertyPartShape",
            "removalshape",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", ""),
        )

        for n in self.propertyEnumerations():
            setattr(obj, n[0], n[1])

        obj.setEditorMode("removalshape", 2)  # hide

        FeatureExtensions.initialize_properties(obj)

    def opSetDefaultValues(self, obj, job):
        obj.Side = "Inside"
        obj.OperationType = "Clearing"
        obj.Tolerance = 0.1
        obj.StepOver = 20
        obj.LiftDistance = 0
        # obj.ProcessHoles = True
        obj.ForceInsideOut = False
        obj.FinishingProfile = True
        obj.Stopped = False
        obj.StopProcessing = False
        obj.HelixAngle = 5
        obj.HelixConeAngle = 0
        obj.HelixDiameterLimit = 0.0
        obj.AdaptiveInputState = ""
        obj.AdaptiveOutputState = ""
        obj.StockToLeave = 0
        obj.KeepToolDownRatio = 3.0
        obj.UseHelixArcs = False
        obj.UseOutline = False
        FeatureExtensions.set_default_property_values(obj, job)

    def opExecute(self, obj):
        """opExecute(obj) ... called whenever the receiver needs to be recalculated.
        See documentation of execute() for a list of base functionality provided.
        Should be overwritten by subclasses."""

        self.pathArray = _get_working_edges(self, obj)
        Execute(self, obj)

    def opOnDocumentRestored(self, obj):
        if not hasattr(obj, "HelixConeAngle"):
            obj.addProperty(
                "App::PropertyAngle",
                "HelixConeAngle",
                "Adaptive",
                "Helix cone angle (degrees)",
            )

        if not hasattr(obj, "UseOutline"):
            obj.addProperty(
                "App::PropertyBool",
                "UseOutline",
                "Adaptive",
                "Uses the outline of the base geometry.",
            )

        if not hasattr(obj, "removalshape"):
            obj.addProperty("Part::PropertyPartShape", "removalshape", "Path", "")
        obj.setEditorMode("removalshape", 2)  # hide

        FeatureExtensions.initialize_properties(obj)


# Eclass


def SetupProperties():
    setup = [
        "Side",
        "OperationType",
        "Tolerance",
        "StepOver",
        "LiftDistance",
        "KeepToolDownRatio",
        "StockToLeave",
        "ForceInsideOut",
        "FinishingProfile",
        "Stopped",
        "StopProcessing",
        "UseHelixArcs",
        "AdaptiveInputState",
        "AdaptiveOutputState",
        "HelixAngle",
        "HelixConeAngle",
        "HelixDiameterLimit",
        "UseOutline",
    ]
    return setup


def Create(name, obj=None, parentJob=None):
    """Create(name) ... Creates and returns a Adaptive operation."""
    if obj is None:
        obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    obj.Proxy = PathAdaptive(obj, name, parentJob)
    return obj



# -*- coding: utf-8 -*-

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Lesser General Public            *
# *   License as published by the Free Software Foundation; either          *
# *   version 2.1 of the License, or (at your option) any later version.    *
# *                                                                         *
# *   This library is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with this library; if not, write to the Free Software   *
# *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
# *   02110-1301  USA                                                       *
# *                                                                         *
# ***************************************************************************

import os

import FreeCAD
import FreeCADGui

from PySide2 import QtWidgets, QtCore

translate = FreeCAD.Qt.translate

try:
    import git

    NO_GIT = False
except Exception:
    NO_GIT = True


class ChangeBranchDialog(QtWidgets.QWidget):

    branch_changed = QtCore.Signal(str)

    def __init__(self, path: os.PathLike, parent=None):
        super().__init__(parent)

        self.ui = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "change_branch.ui")
        )
        self.ui.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)

        self.item_filter = ChangeBranchDialogFilter()
        self.ui.tableView.setModel(self.item_filter)

        self.item_model = ChangeBranchDialogModel(path, self)
        self.item_filter.setSourceModel(self.item_model)
        self.ui.tableView.sortByColumn(
            4, QtCore.Qt.DescendingOrder
        )  # Default to sorting by remote last-changed date

        # Figure out what row gets selected:
        row = 0
        current_ref = self.item_model.repo.head.ref
        selection_model = self.ui.tableView.selectionModel()
        for ref in self.item_model.refs:
            if ref == current_ref:
                index = self.item_filter.mapFromSource(self.item_model.index(row, 0))
                selection_model.select(index, QtCore.QItemSelectionModel.ClearAndSelect)
                selection_model.select(
                    index.siblingAtColumn(1), QtCore.QItemSelectionModel.Select
                )
                selection_model.select(
                    index.siblingAtColumn(2), QtCore.QItemSelectionModel.Select
                )
                selection_model.select(
                    index.siblingAtColumn(3), QtCore.QItemSelectionModel.Select
                )
                selection_model.select(
                    index.siblingAtColumn(4), QtCore.QItemSelectionModel.Select
                )
                break
            row += 1

        # Make sure the column widths are OK:
        header = self.ui.tableView.horizontalHeader()
        header.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

    def exec(self):
        if self.ui.exec() == QtWidgets.QDialog.Accepted:

            selection = self.ui.tableView.selectedIndexes()
            index = self.item_filter.mapToSource(selection[0])
            ref = self.item_model.data(index, ChangeBranchDialogModel.RefAccessRole)

            if ref == self.item_model.repo.head.ref:
                # This is the one we are already on... just return
                return

            result = QtWidgets.QMessageBox.critical(
                self,
                translate("AddonsInstaller", "DANGER: Developer feature"),
                translate(
                    "AddonsInstaller",
                    "DANGER: Switching branches is intended for developers and beta testers, and may result in broken, non-backwards compatible documents, instability, crashes, and/or the premature heat death of the universe. Are you sure you want to continue?",
                ),
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.Cancel,
                QtWidgets.QMessageBox.Cancel,
            )
            if result == QtWidgets.QMessageBox.Cancel:
                return
            if self.item_model.repo.is_dirty():
                result = QtWidgets.QMessageBox.critical(
                    self,
                    translate("AddonsInstaller", "There are local changes"),
                    translate(
                        "AddonsInstaller",
                        "WARNING: This repo has uncommitted local changes. Are you sure you want to change branches (bringing the changes with you)?",
                    ),
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.Cancel,
                    QtWidgets.QMessageBox.Cancel,
                )
                if result == QtWidgets.QMessageBox.Cancel:
                    return

            if isinstance(ref, git.TagReference):
                # Detach the head
                self.item_model.repo.head.reference = ref
                self.item_model.repo.head.reset(index=True, working_tree=True)
            elif isinstance(ref, git.RemoteReference):
                # Set up a local tracking branch
                slash_index = ref.name.find("/")
                if slash_index != -1:
                    local_name = ref.name[slash_index + 1 :]
                else:
                    local_name = ref.name
                self.item_model.repo.create_head(local_name, ref)
                self.item_model.repo.heads[local_name].set_tracking_branch(ref)
                self.item_model.repo.heads[local_name].checkout()
            else:
                # It's already a local branch, just check it out
                ref.checkout()

            self.branch_changed.emit(ref.name)


class ChangeBranchDialogModel(QtCore.QAbstractTableModel):

    refs = []
    display_data = []
    DataSortRole = QtCore.Qt.UserRole
    RefAccessRole = QtCore.Qt.UserRole + 1

    def __init__(self, path: os.PathLike, parent=None) -> None:
        super().__init__(parent)
        self.repo = git.Repo(path)

        self.refs = []
        tracking_refs = []
        for ref in self.repo.refs:
            row = ["", None, None, None, None]
            if "HEAD" in ref.name:
                continue
            if isinstance(ref, git.RemoteReference):
                if ref.name in tracking_refs:
                    # Already seen, it's the remote part of a remote tracking branch
                    continue
                else:
                    # Just a remote branch, not tracking:
                    row[0] = translate("AddonsInstaller", "Branch", "git terminology")
                    row[2] = ref.name
                    if hasattr(ref, "commit") and hasattr(ref.commit, "committed_date"):
                        row[4] = ref.commit.committed_date
                    else:
                        row[4] = ref.log_entry(0).time[0]
            elif isinstance(ref, git.TagReference):
                # Tags are simple, there is no tracking to worry about
                row[0] = translate("AddonsInstaller", "Tag", "git terminology")
                row[1] = ref.name
                row[3] = ref.commit.committed_date
            elif isinstance(ref, git.Head):
                if hasattr(ref, "tracking_branch") and ref.tracking_branch():
                    # This local branch tracks a remote: we have all five pieces of data...
                    row[0] = translate("AddonsInstaller", "Branch", "git terminology")
                    row[1] = ref.name
                    row[2] = ref.tracking_branch().name
                    row[3] = ref.commit.committed_date
                    row[4] = ref.tracking_branch().commit.committed_date
                    tracking_refs.append(ref.tracking_branch().name)
                else:
                    # Just a local branch, no remote tracking:
                    row[0] = translate("AddonsInstaller", "Branch", "git terminology")
                    row[1] = ref.name
                    if hasattr(ref, "commit") and hasattr(ref.commit, "committed_date"):
                        row[3] = ref.commit.committed_date
                    else:
                        row[3] = ref.log_entry(0).time[0]
            else:
                continue

            self.display_data.append(row.copy())
            self.refs.append(ref)

    def rowCount(self, parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        if parent.isValid():
            return 0
        return len(self.refs)

    def columnCount(self, parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        if parent.isValid():
            return 0
        return 5  # Type, local name, tracking name, local update, and remote update

    def data(self, index: QtCore.QModelIndex, role: int = QtCore.Qt.DisplayRole):
        if not index.isValid():
            return None
        row = index.row()
        column = index.column()
        if role == QtCore.Qt.ToolTipRole:
            tooltip = ""
            # TODO: What should the tooltip be for these items? Last commit message?
            return tooltip
        elif role == QtCore.Qt.DisplayRole:
            dd = self.display_data[row]
            if column == 3 or column == 4:
                if dd[column] is not None:
                    qdate = QtCore.QDateTime.fromTime_t(dd[column])
                    return QtCore.QLocale().toString(qdate, QtCore.QLocale.ShortFormat)
            elif column < len(dd):
                return dd[column]
            else:
                return None
        elif role == ChangeBranchDialogModel.DataSortRole:
            if column == 0:
                if self.refs[row] in self.repo.heads:
                    return 0
                else:
                    return 1
            elif column < len(self.display_data[row]):
                return self.display_data[row][column]
            else:
                return None
        elif role == ChangeBranchDialogModel.RefAccessRole:
            return self.refs[row]

    def headerData(
        self,
        section: int,
        orientation: QtCore.Qt.Orientation,
        role: int = QtCore.Qt.DisplayRole,
    ):
        if orientation == QtCore.Qt.Vertical:
            return None
        if role != QtCore.Qt.DisplayRole:
            return None
        if section == 0:
            return translate(
                "AddonsInstaller",
                "Kind",
                "Table header for git ref type (e.g. either Tag or Branch)",
            )
        elif section == 1:
            return translate(
                "AddonsInstaller", "Local name", "Table header for git ref name"
            )
        elif section == 2:
            return translate(
                "AddonsInstaller",
                "Tracking",
                "Table header for git remote tracking branch name name",
            )
        elif section == 3:
            return translate(
                "AddonsInstaller",
                "Local updated",
                "Table header for git update time of local branch",
            )
        elif section == 4:
            return translate(
                "AddonsInstaller",
                "Remote updated",
                "Table header for git update time of remote branch",
            )
        else:
            return None


class ChangeBranchDialogFilter(QtCore.QSortFilterProxyModel):
    def lessThan(self, left: QtCore.QModelIndex, right: QtCore.QModelIndex):
        leftData = self.sourceModel().data(left, ChangeBranchDialogModel.DataSortRole)
        rightData = self.sourceModel().data(right, ChangeBranchDialogModel.DataSortRole)
        if leftData is None or rightData is None:
            if rightData is not None:
                return True
            else:
                return False
        return leftData < rightData



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# test script for TechDraw module
# creates a page and 2 views
# adds 1 length dimension to view1
# adds 1 radius dimension to view2
from __future__ import print_function

import FreeCAD
from FreeCAD import Units
import os


def DVBalloonTest():
    path = os.path.dirname(os.path.abspath(__file__))
    print("TDBalloon path: " + path)
    templateFileSpec = path + "/TestTemplate.svg"

    FreeCAD.newDocument("TDBalloon")
    FreeCAD.setActiveDocument("TDBalloon")
    FreeCAD.ActiveDocument = FreeCAD.getDocument("TDBalloon")

    # make source feature
    FreeCAD.ActiveDocument.addObject("Part::Box", "Box")
    FreeCAD.ActiveDocument.addObject("Part::Sphere", "Sphere")

    # make a page
    page = FreeCAD.ActiveDocument.addObject("TechDraw::DrawPage", "Page")
    FreeCAD.ActiveDocument.addObject("TechDraw::DrawSVGTemplate", "Template")
    FreeCAD.ActiveDocument.Template.Template = templateFileSpec
    FreeCAD.ActiveDocument.Page.Template = FreeCAD.ActiveDocument.Template
    page.Scale = 5.0
    # page.ViewObject.show()  # unit tests run in console mode

    # make Views
    view1 = FreeCAD.ActiveDocument.addObject("TechDraw::DrawViewPart", "View")
    FreeCAD.ActiveDocument.View.Source = [FreeCAD.ActiveDocument.Box]
    page.addView(view1)
    view1.X = Units.Quantity(30.0, Units.Length)
    view1.Y = Units.Quantity(150.0, Units.Length)
    view2 = FreeCAD.activeDocument().addObject("TechDraw::DrawViewPart", "View001")
    FreeCAD.activeDocument().View001.Source = [FreeCAD.activeDocument().Sphere]
    page.addView(view2)
    view2.X = Units.Quantity(220.0, Units.Length)
    view2.Y = Units.Quantity(150.0, Units.Length)
    FreeCAD.ActiveDocument.recompute()

    print("Place balloon")
    balloon1 = FreeCAD.ActiveDocument.addObject("TechDraw::DrawViewBalloon", "Balloon1")
    balloon1.SourceView = view1
    # balloon1.OriginIsSet = 1  # OriginIsSet property removed March 2020
    balloon1.OriginX = view1.X + Units.Quantity(20.0, Units.Length)
    balloon1.OriginY = view1.Y + Units.Quantity(20.0, Units.Length)
    balloon1.Text = "1"
    balloon1.Y = balloon1.OriginX + Units.Quantity(20.0, Units.Length)
    balloon1.X = balloon1.OriginY + Units.Quantity(20.0, Units.Length)

    print("adding balloon1 to page")
    page.addView(balloon1)

    balloon2 = FreeCAD.ActiveDocument.addObject("TechDraw::DrawViewBalloon", "Balloon2")
    balloon2.SourceView = view2
    # balloon2.OriginIsSet = 1
    balloon2.OriginX = view2.X + Units.Quantity(20.0, Units.Length)
    balloon2.OriginY = view2.Y + Units.Quantity(20.0, Units.Length)
    balloon2.Text = "2"
    balloon2.Y = balloon2.OriginX + Units.Quantity(20.0, Units.Length)
    balloon2.X = balloon2.OriginY + Units.Quantity(20.0, Units.Length)

    print("adding balloon2 to page")
    page.addView(balloon2)

    FreeCAD.ActiveDocument.recompute()

    rc = False
    if ("Up-to-date" in balloon1.State) and ("Up-to-date" in balloon2.State):
        rc = True

    FreeCAD.closeDocument("TDBalloon")
    return rc


if __name__ == "__main__":
    DVBalloonTest()



#***************************************************************************
#*   Copyright (c) 2011 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************

import unittest

import FreeCAD

App = FreeCAD

class TestBoolean(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartDesignTestBoolean")

    def testBooleanFuseCase(self):
        self.Body = self.Doc.addObject('PartDesign::Body','Body')
        self.Box = self.Doc.addObject('PartDesign::AdditiveBox','Box')
        self.Box.Length=10
        self.Box.Width=10
        self.Box.Height=10
        self.Body.addObject(self.Box)
        self.Doc.recompute()
        self.Body001 = self.Doc.addObject('PartDesign::Body','Body001')
        self.Box001 = self.Doc.addObject('PartDesign::AdditiveBox','Box001')
        self.Box001.Length=10
        self.Box001.Width=10
        self.Box001.Height=10
        self.Box001.Placement.Base = App.Vector(-5,0,0)
        self.Body001.addObject(self.Box001)
        self.Doc.recompute()
        self.BooleanFuse = self.Doc.addObject('PartDesign::Boolean','BooleanFuse')
        self.Body001.addObject(self.BooleanFuse)
        self.Doc.recompute()
        self.BooleanFuse.setObjects([self.Body,])
        self.BooleanFuse.Type = 0
        self.Doc.recompute()
        self.assertAlmostEqual(self.BooleanFuse.Shape.Volume, 1500)
        
    def testBooleanCutCase(self):
        self.Body = self.Doc.addObject('PartDesign::Body','Body')
        self.Box = self.Doc.addObject('PartDesign::AdditiveBox','Box')
        self.Box.Length=10
        self.Box.Width=10
        self.Box.Height=10
        self.Body.addObject(self.Box)
        self.Doc.recompute()
        self.Body001 = self.Doc.addObject('PartDesign::Body','Body001')
        self.Box001 = self.Doc.addObject('PartDesign::AdditiveBox','Box001')
        self.Box001.Length=10
        self.Box001.Width=10
        self.Box001.Height=10
        self.Box001.Placement.Base = App.Vector(-5,0,0)
        self.Body001.addObject(self.Box001)
        self.Doc.recompute()
        self.BooleanCut = self.Doc.addObject('PartDesign::Boolean','BooleanCut')
        self.Body001.addObject(self.BooleanCut)
        self.Doc.recompute()
        self.BooleanCut.setObjects([self.Body,])
        self.BooleanCut.Type = 1
        self.Doc.recompute()
        self.assertAlmostEqual(self.BooleanCut.Shape.Volume, 500)

    def testBooleanCommonCase(self):
        self.Body = self.Doc.addObject('PartDesign::Body','Body')
        self.Box = self.Doc.addObject('PartDesign::AdditiveBox','Box')
        self.Box.Length=10
        self.Box.Width=10
        self.Box.Height=10
        self.Body.addObject(self.Box)
        self.Doc.recompute()
        self.Body001 = self.Doc.addObject('PartDesign::Body','Body001')
        self.Box001 = self.Doc.addObject('PartDesign::AdditiveBox','Box001')
        self.Box001.Length=10
        self.Box001.Width=10
        self.Box001.Height=10
        self.Box001.Placement.Base = App.Vector(-5,0,0)
        self.Body001.addObject(self.Box001)
        self.Doc.recompute()
        self.BooleanCommon = self.Doc.addObject('PartDesign::Boolean','BooleanCommon')
        self.Body001.addObject(self.BooleanCommon)
        self.Doc.recompute()
        self.BooleanCommon.setObjects([self.Body,])
        self.BooleanCommon.Type = 2
        self.Doc.recompute()
        self.assertAlmostEqual(self.BooleanCommon.Shape.Volume, 500)

    def tearDown(self):
        #closing doc
        FreeCAD.closeDocument("PartDesignTestBoolean")
        #print ("omit closing document for debugging")

